# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""The ExpressionManager is used to create expressions.

All objects are memoized so that two syntactically equivalent expressions
are represented by the same object.
"""


import unified_planning as up
import unified_planning.model.types
from unified_planning.model.operators import OperatorKind
from unified_planning.exceptions import UPTypeError, UPExpressionDefinitionError
from fractions import Fraction
from typing import Iterable, List, Union, Dict, Tuple

Expression = Union[
    "up.model.fnode.FNode",
    "up.model.fluent.Fluent",
    "up.model.object.Object",
    "up.model.parameter.Parameter",
    "up.model.variable.Variable",
    "up.model.timing.Timing",
    bool,
    int,
    float,
    Fraction,
]
BoolExpression = Union[
    "up.model.fnode.FNode",
    "up.model.fluent.Fluent",
    "up.model.parameter.Parameter",
    bool,
]
ConstantExpression = Union[
    "up.model.fnode.FNode", "up.model.object.Object", bool, int, float, Fraction
]


class ExpressionManager(object):
    """ExpressionManager is responsible for the creation of all expressions."""

    def __init__(self, env: "up.environment.Environment"):
        self.env = env
        self.expressions: Dict[
            "up.model.fnode.FNodeContent", "up.model.fnode.FNode"
        ] = {}
        self._next_free_id = 1

        self.true_expression = self.create_node(
            node_type=OperatorKind.BOOL_CONSTANT, args=tuple(), payload=True
        )
        self.false_expression = self.create_node(
            node_type=OperatorKind.BOOL_CONSTANT, args=tuple(), payload=False
        )
        return

    def _polymorph_args_to_tuple(
        self, *args: Union[Expression, Iterable[Expression]]
    ) -> Tuple[Expression, ...]:
        """Helper function to return a tuple of arguments from args.
        This function is used to allow N-ary operators to express their arguments
        both as a list of arguments or as a tuple of arguments: e.g.,
           And([a,b,c]) and And(a,b,c)
        are both valid, and they are converted into a tuple (a,b,c)"""

        res = []
        for p in args:
            if isinstance(p, Iterable):
                res.extend(list(p))
            else:
                res.append(p)
        return tuple(res)

    def auto_promote(
        self, *args: Union[Expression, Iterable[Expression]]
    ) -> List["up.model.fnode.FNode"]:
        tuple_args = self._polymorph_args_to_tuple(*args)
        res = []
        for e in tuple_args:
            if isinstance(e, up.model.fluent.Fluent):
                assert (
                    e.environment == self.env
                ), "Fluent has a different environment of the expression manager"
                res.append(self.FluentExp(e))
            elif isinstance(e, up.model.parameter.Parameter):
                assert (
                    e.environment == self.env
                ), "Parameter has a different environment of the expression manager"
                res.append(self.ParameterExp(e))
            elif isinstance(e, up.model.variable.Variable):
                assert (
                    e.environment == self.env
                ), "Variable has a different environment of the expression manager"
                res.append(self.VariableExp(e))
            elif isinstance(e, up.model.object.Object):
                assert (
                    e.environment == self.env
                ), "Object has a different environment of the expression manager"
                res.append(self.ObjectExp(e))
            elif isinstance(e, up.model.timing.Timing):
                res.append(self.TimingExp(e))
            elif isinstance(e, bool):
                res.append(self.Bool(e))
            elif isinstance(e, int):
                res.append(self.Int(e))
            elif isinstance(e, float):
                res.append(self.Real(Fraction(e)))
            elif isinstance(e, Fraction):
                res.append(self.Real(e))
            else:
                assert (
                    e.environment == self.env
                ), "Expression has a different environment of the expression manager"
                res.append(e)
        return res

    def create_node(
        self,
        node_type: OperatorKind,
        args: Iterable["up.model.fnode.FNode"],
        payload: Union[
            "up.model.fluent.Fluent",
            "up.model.object.Object",
            "up.model.parameter.Parameter",
            "up.model.variable.Variable",
            "up.model.timing.Timing",
            bool,
            int,
            Fraction,
            Tuple["up.model.variable.Variable", ...],
        ] = None,
    ) -> "up.model.fnode.FNode":
        content = up.model.fnode.FNodeContent(node_type, args, payload)
        if content in self.expressions:
            return self.expressions[content]
        else:
            assert all(
                a.environment == self.env for a in args
            ), "2 FNode in the same expression have different environments"
            n = up.model.fnode.FNode(content, self._next_free_id, self.env)
            self._next_free_id += 1
            self.expressions[content] = n
            self.env.type_checker.get_type(n)
            return n

    def And(
        self, *args: Union[BoolExpression, Iterable[BoolExpression]]
    ) -> "up.model.fnode.FNode":
        """Returns a conjunction of terms.
        This function has polimorphic n-arguments:
          - And(a,b,c)
          - And([a,b,c])
        Restriction: Arguments must be boolean
        """
        tuple_args = tuple(self.auto_promote(*args))

        if len(tuple_args) == 0:
            return self.TRUE()
        elif len(tuple_args) == 1:
            return tuple_args[0]
        else:
            return self.create_node(node_type=OperatorKind.AND, args=tuple_args)

    def Or(
        self, *args: Union[BoolExpression, Iterable[BoolExpression]]
    ) -> "up.model.fnode.FNode":
        """Returns an disjunction of terms.
        This function has polimorphic n-arguments:
          - Or(a,b,c)
          - Or([a,b,c])
        Restriction: Arguments must be boolean
        """
        tuple_args = tuple(self.auto_promote(*args))

        if len(tuple_args) == 0:
            return self.FALSE()
        elif len(tuple_args) == 1:
            return tuple_args[0]
        else:
            return self.create_node(node_type=OperatorKind.OR, args=tuple_args)

    def Not(self, expression: BoolExpression) -> "up.model.fnode.FNode":
        """Creates an expression of the form:
                not expression
        Restriction: expression must be of boolean type
        """
        (expression,) = self.auto_promote(expression)
        if expression.is_not():
            return expression.arg(0)
        return self.create_node(node_type=OperatorKind.NOT, args=(expression,))

    def Implies(
        self, left: BoolExpression, right: BoolExpression
    ) -> "up.model.fnode.FNode":
        """Creates an expression of the form:
            left -> right
        Restriction: Left and Right must be of boolean type
        """
        left, right = self.auto_promote(left, right)
        return self.create_node(node_type=OperatorKind.IMPLIES, args=(left, right))

    def Iff(
        self, left: BoolExpression, right: BoolExpression
    ) -> "up.model.fnode.FNode":
        """Creates an expression of the form:
            left <-> right
        Restriction: Left and Right must be of boolean type
        """
        left, right = self.auto_promote(left, right)
        return self.create_node(node_type=OperatorKind.IFF, args=(left, right))

    def Exists(
        self, expression: BoolExpression, *vars: "up.model.variable.Variable"
    ) -> "up.model.fnode.FNode":
        """Creates an expression of the form:
            Exists (var[0]... var[n]) | expression
        Restriction: expression must be of boolean type and
                    vars must be of 'up.Variable' type
        """
        expressions = tuple(self.auto_promote(expression))
        if len(vars) == 0:
            raise UPExpressionDefinitionError(
                f"Exists of expression: {str(expression)} must be created with at least one variable, otherwise it is not needed."
            )
        for v in vars:
            if not isinstance(v, up.model.variable.Variable):
                raise UPTypeError("Expecting 'up.Variable', got %s", type(v))
        return self.create_node(
            node_type=OperatorKind.EXISTS, args=expressions, payload=vars
        )

    def Forall(
        self, expression: BoolExpression, *vars: "up.model.variable.Variable"
    ) -> "up.model.fnode.FNode":
        """Creates an expression of the form:
            Forall (var[0]... var[n]) | expression
        Restriction: expression must be of boolean type and
                    vars must be of 'up.Variable' type
        """
        expressions = tuple(self.auto_promote(expression))
        if len(vars) == 0:
            raise UPExpressionDefinitionError(
                f"Forall of expression: {str(expression)} must be created with at least one variable, otherwise it is not needed."
            )
        for v in vars:
            if not isinstance(v, up.model.variable.Variable):
                raise UPTypeError("Expecting 'up.Variable', got %s", type(v))
        return self.create_node(
            node_type=OperatorKind.FORALL, args=expressions, payload=vars
        )

    def FluentExp(
        self, fluent: "up.model.fluent.Fluent", params: Tuple[Expression, ...] = tuple()
    ) -> "up.model.fnode.FNode":
        """Creates an expression for the given fluent and parameters.
        Restriction: parameters type must be compatible with the fluent signature
        """
        assert fluent.arity == len(params)
        assert fluent.environment == self.env
        params_exp = self.auto_promote(*params)
        return self.create_node(
            node_type=OperatorKind.FLUENT_EXP, args=tuple(params_exp), payload=fluent
        )

    def ParameterExp(
        self, param: "up.model.parameter.Parameter"
    ) -> "up.model.fnode.FNode":
        """Returns an expression for the given action parameter."""
        return self.create_node(
            node_type=OperatorKind.PARAM_EXP, args=tuple(), payload=param
        )

    def VariableExp(self, var: "up.model.variable.Variable") -> "up.model.fnode.FNode":
        """Returns an expression for the given variable."""
        assert var.environment == self.env
        return self.create_node(
            node_type=OperatorKind.VARIABLE_EXP, args=tuple(), payload=var
        )

    def ObjectExp(self, obj: "up.model.object.Object") -> "up.model.fnode.FNode":
        """Returns an expression for the given object."""
        assert obj.environment == self.env
        return self.create_node(
            node_type=OperatorKind.OBJECT_EXP, args=tuple(), payload=obj
        )

    def TimingExp(self, obj: "up.model.timing.Timing") -> "up.model.fnode.FNode":
        """Returns an expression for the given timing."""
        return self.create_node(
            node_type=OperatorKind.TIMING_EXP, args=tuple(), payload=obj
        )

    def TRUE(self) -> "up.model.fnode.FNode":
        """Return the boolean constant True."""
        return self.true_expression

    def FALSE(self) -> "up.model.fnode.FNode":
        """Return the boolean constant False."""
        return self.false_expression

    def Bool(self, value: bool) -> "up.model.fnode.FNode":
        """Return a boolean constant."""
        if type(value) != bool:
            raise UPTypeError("Expecting bool, got %s" % type(value))

        if value:
            return self.true_expression
        else:
            return self.false_expression

    def Int(self, value: int) -> "up.model.fnode.FNode":
        """Return an int constant."""
        if type(value) != int:
            raise UPTypeError("Expecting int, got %s" % type(value))
        return self.create_node(
            node_type=OperatorKind.INT_CONSTANT, args=tuple(), payload=value
        )

    def Real(self, value: Fraction) -> "up.model.fnode.FNode":
        """Return a real constant."""
        if type(value) != Fraction:
            raise UPTypeError("Expecting Fraction, got %s" % type(value))
        return self.create_node(
            node_type=OperatorKind.REAL_CONSTANT, args=tuple(), payload=value
        )

    def Plus(
        self, *args: Union[Expression, Iterable[Expression]]
    ) -> "up.model.fnode.FNode":
        """Creates an expression of the form:
        args[0] + ... + args[n]
        """
        tuple_args = tuple(self.auto_promote(*args))

        if len(tuple_args) == 0:
            return self.Int(0)
        elif len(tuple_args) == 1:
            return tuple_args[0]
        else:
            return self.create_node(node_type=OperatorKind.PLUS, args=tuple_args)

    def Minus(self, left: Expression, right: Expression) -> "up.model.fnode.FNode":
        """Creates an expression of the form: left - right."""
        left, right = self.auto_promote(left, right)
        return self.create_node(node_type=OperatorKind.MINUS, args=(left, right))

    def Times(
        self, *args: Union[Expression, Iterable[Expression]]
    ) -> "up.model.fnode.FNode":
        """Creates an expression of the form:
        args[0] * ... * args[n]
        """
        tuple_args = tuple(self.auto_promote(*args))

        if len(tuple_args) == 0:
            return self.Int(1)
        elif len(tuple_args) == 1:
            return tuple_args[0]
        else:
            return self.create_node(node_type=OperatorKind.TIMES, args=tuple_args)

    def Div(self, left: Expression, right: Expression) -> "up.model.fnode.FNode":
        """Creates an expression of the form: left / right."""
        left, right = self.auto_promote(left, right)
        return self.create_node(node_type=OperatorKind.DIV, args=(left, right))

    def LE(self, left: Expression, right: Expression) -> "up.model.fnode.FNode":
        """Creates an expression of the form: left <= right."""
        left, right = self.auto_promote(left, right)
        return self.create_node(node_type=OperatorKind.LE, args=(left, right))

    def GE(self, left: Expression, right: Expression) -> "up.model.fnode.FNode":
        """Creates an expression of the form: left >= right."""
        left, right = self.auto_promote(left, right)
        return self.create_node(node_type=OperatorKind.LE, args=(right, left))

    def LT(self, left: Expression, right: Expression) -> "up.model.fnode.FNode":
        """Creates an expression of the form: left < right."""
        left, right = self.auto_promote(left, right)
        return self.create_node(node_type=OperatorKind.LT, args=(left, right))

    def GT(self, left: Expression, right: Expression) -> "up.model.fnode.FNode":
        """Creates an expression of the form: left > right."""
        left, right = self.auto_promote(left, right)
        return self.create_node(node_type=OperatorKind.LT, args=(right, left))

    def Equals(self, left: Expression, right: Expression) -> "up.model.fnode.FNode":
        """Creates an expression of the form: left == right."""
        left, right = self.auto_promote(left, right)
        return self.create_node(node_type=OperatorKind.EQUALS, args=(left, right))



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import unified_planning as up
import unified_planning.environment
import unified_planning.engines as engines
import unified_planning.engines.mixins as mixins
import unified_planning.model.walkers as walkers
from unified_planning.model import (
    AbstractProblem,
    Problem,
    ProblemKind,
    COWState,
    UPCOWState,
)
from unified_planning.engines.results import (
    ValidationResult,
    ValidationResultStatus,
    LogMessage,
    LogLevel,
)
from unified_planning.engines.sequential_simulator import (
    SequentialSimulator,
    InstantaneousEvent,
)
from unified_planning.plans import SequentialPlan, PlanKind


class SequentialPlanValidator(engines.engine.Engine, mixins.PlanValidatorMixin):
    """Performs plan validation."""

    def __init__(self, **options):
        self._env: "unified_planning.environment.Environment" = (
            unified_planning.environment.get_env(options.get("env", None))
        )
        self.manager = self._env.expression_manager
        self._substituter = walkers.Substituter(self._env)

    @property
    def name(self):
        return "sequential_plan_validator"

    @staticmethod
    def supports_plan(plan_kind: "up.plans.PlanKind") -> bool:
        return plan_kind == PlanKind.SEQUENTIAL_PLAN

    @staticmethod
    def supported_kind() -> ProblemKind:
        supported_kind = ProblemKind()
        supported_kind.set_problem_class("ACTION_BASED")
        supported_kind.set_typing("FLAT_TYPING")
        supported_kind.set_typing("HIERARCHICAL_TYPING")
        supported_kind.set_numbers("CONTINUOUS_NUMBERS")
        supported_kind.set_numbers("DISCRETE_NUMBERS")
        supported_kind.set_conditions_kind("NEGATIVE_CONDITIONS")
        supported_kind.set_conditions_kind("DISJUNCTIVE_CONDITIONS")
        supported_kind.set_conditions_kind("EQUALITY")
        supported_kind.set_conditions_kind("EXISTENTIAL_CONDITIONS")
        supported_kind.set_conditions_kind("UNIVERSAL_CONDITIONS")
        supported_kind.set_effects_kind("CONDITIONAL_EFFECTS")
        supported_kind.set_effects_kind("INCREASE_EFFECTS")
        supported_kind.set_effects_kind("DECREASE_EFFECTS")
        supported_kind.set_fluents_type("NUMERIC_FLUENTS")
        supported_kind.set_fluents_type("OBJECT_FLUENTS")
        return supported_kind

    @staticmethod
    def supports(problem_kind):
        return problem_kind <= SequentialPlanValidator.supported_kind()

    def _validate(
        self, problem: "AbstractProblem", plan: "unified_planning.plans.Plan"
    ) -> "up.engines.results.ValidationResult":
        """Returns True if and only if the plan given in input is a valid plan for the problem given in input.
        This means that from the initial state of the problem, by following the plan, you can reach the
        problem goal. Otherwise False is returned."""
        assert isinstance(plan, SequentialPlan)
        assert isinstance(problem, Problem)
        simulator = SequentialSimulator(problem)
        current_state: "COWState" = UPCOWState(problem.initial_values)
        count = 0  # used for better error indexing
        for ai in plan.actions:
            action = ai.action
            assert isinstance(action, unified_planning.model.InstantaneousAction)
            count = count + 1
            events = simulator.get_events(action, ai.actual_parameters)
            assert len(events) < 2, f"{str(ai)} + {len(events)}"
            for event in events:
                if not simulator.is_applicable(event, current_state):
                    error = f"Preconditions {event.conditions} of {str(count)}-th action instance {str(ai)} are not satisfied."
                    logs = [LogMessage(LogLevel.ERROR, error)]
                    return ValidationResult(
                        ValidationResultStatus.INVALID, self.name, logs
                    )
                current_state = simulator.apply_unsafe(event, current_state)
        unsatisfied_goals = simulator.get_unsatisfied_goals(current_state)
        if len(unsatisfied_goals) == 0:
            return ValidationResult(ValidationResultStatus.VALID, self.name, [])
        else:
            error = f"Goals {unsatisfied_goals} are not satisfied by the plan."
            logs = [LogMessage(LogLevel.ERROR, error)]
            return ValidationResult(ValidationResultStatus.INVALID, self.name, logs)



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unified_planning as up
from unified_planning.shortcuts import *
from unified_planning.test import TestCase, skipIfEngineNotAvailable


class TestSimulatedEffects(TestCase):
    def setUp(self):
        TestCase.setUp(self)

    @skipIfEngineNotAvailable("tamer")
    def test_basic(self):
        x = Fluent("x")
        a = InstantaneousAction("a")
        a.add_precondition(Not(x))

        def fun(problem, state, actual_params):
            if state.get_value(FluentExp(x)).is_false():
                return [TRUE()]
            else:
                return [FALSE()]

        a.set_simulated_effect(SimulatedEffect([FluentExp(x)], fun))
        problem = Problem("basic_with_simulated_effects")
        problem.add_fluent(x)
        problem.add_action(a)
        problem.set_initial_value(x, False)
        problem.add_goal(x)

        self.assertTrue(problem.kind.has_simulated_effects())

        with OneshotPlanner(name="tamer") as planner:
            self.assertNotEqual(planner, None)
            res = planner.solve(problem)
            plan = res.plan
            self.assertEqual(len(plan.actions), 1)

    @skipIfEngineNotAvailable("tamer")
    def test_with_parameters(self):
        Location = UserType("Location")
        Robot = UserType("Robot")
        at = Fluent("at", Location, robot=Robot)
        battery_charge = Fluent("battery_charge", IntType(0, 100), robot=Robot)
        move = InstantaneousAction("move", robot=Robot, l_from=Location, l_to=Location)
        robot = move.parameter("robot")
        l_from = move.parameter("l_from")
        l_to = move.parameter("l_to")
        move.add_precondition(Equals(at(robot), l_from))
        move.add_precondition(GE(battery_charge(robot), 10))
        move.add_precondition(Not(Equals(l_from, l_to)))
        move.add_effect(at(robot), l_to)

        def fun(problem, state, actual_params):
            value = state.get_value(
                battery_charge(actual_params.get(robot))
            ).constant_value()
            return [Int(value - 10)]

        move.set_simulated_effect(SimulatedEffect([battery_charge(robot)], fun))
        l1 = Object("l1", Location)
        l2 = Object("l2", Location)
        r1 = Object("r1", Robot)
        problem = Problem("robot_with_simulated_effects")
        problem.add_fluent(at)
        problem.add_fluent(battery_charge)
        problem.add_action(move)
        problem.add_object(l1)
        problem.add_object(l2)
        problem.add_object(r1)
        problem.set_initial_value(at(r1), l1)
        problem.set_initial_value(battery_charge(r1), 100)
        problem.add_goal(Equals(at(r1), l2))

        self.assertTrue(problem.kind.has_simulated_effects())

        with OneshotPlanner(name="tamer", params={"heuristic": "hff"}) as planner:
            self.assertNotEqual(planner, None)
            res = planner.solve(problem)
            plan = res.plan
            self.assertEqual(len(plan.actions), 1)

    @skipIfEngineNotAvailable("tamer")
    def test_temporal_basic(self):
        x = Fluent("x")
        y = Fluent("y")
        a = DurativeAction("a")
        a.set_fixed_duration(1)
        a.add_condition(StartTiming(), Not(x))
        a.add_condition(StartTiming(), y)

        def fun(problem, state, actual_params):
            return [TRUE(), FALSE()]

        a.set_simulated_effect(
            EndTiming(), SimulatedEffect([FluentExp(x), FluentExp(y)], fun)
        )
        problem = Problem("temporal_basic_with_simulated_effects")
        problem.add_fluent(x)
        problem.add_fluent(y)
        problem.add_action(a)
        problem.set_initial_value(x, False)
        problem.set_initial_value(y, True)
        problem.add_goal(And(x, Not(y)))

        self.assertTrue(problem.kind.has_simulated_effects())

        with OneshotPlanner(problem_kind=problem.kind) as planner:
            self.assertNotEqual(planner, None)
            res = planner.solve(problem)
            plan = res.plan
            self.assertEqual(len(plan.timed_actions), 1)



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import importlib
import sys
import os
import inspect
import configparser
import unified_planning as up
from unified_planning.environment import Environment
from unified_planning.model import ProblemKind
from unified_planning.plans import PlanKind
from unified_planning.engines.mixins.oneshot_planner import OptimalityGuarantee
from unified_planning.engines.mixins.compiler import CompilationKind, CompilerMixin
from unified_planning.engines.mixins.oneshot_planner import OneshotPlannerMixin
from unified_planning.engines.mixins.plan_validator import PlanValidatorMixin
from typing import IO, Dict, Tuple, Optional, List, Union, Type, cast
from pathlib import PurePath


DEFAULT_ENGINES = {
    "fast-downward": ("up_fast_downward", "FastDownwardPDDLPlanner"),
    "fast-downward-opt": ("up_fast_downward", "FastDownwardOptimalPDDLPlanner"),
    "pyperplan": ("up_pyperplan.engine", "EngineImpl"),
    "enhsp": ("up_enhsp.enhsp_planner", "ENHSPSatEngine"),
    "enhsp-opt": ("up_enhsp.enhsp_planner", "ENHSPOptEngine"),
    "tamer": ("up_tamer.engine", "EngineImpl"),
    "sequential_plan_validator": (
        "unified_planning.engines.plan_validator",
        "SequentialPlanValidator",
    ),
    "sequential_simulator": (
        "unified_planning.engines.sequential_simulator",
        "SequentialSimulator",
    ),
    "up_conditional_effects_remover": (
        "unified_planning.engines.compilers.conditional_effects_remover",
        "ConditionalEffectsRemover",
    ),
    "up_disjunctive_conditions_remover": (
        "unified_planning.engines.compilers.disjunctive_conditions_remover",
        "DisjunctiveConditionsRemover",
    ),
    "up_negative_conditions_remover": (
        "unified_planning.engines.compilers.negative_conditions_remover",
        "NegativeConditionsRemover",
    ),
    "up_quantifiers_remover": (
        "unified_planning.engines.compilers.quantifiers_remover",
        "QuantifiersRemover",
    ),
    "tarski_grounder": (
        "unified_planning.engines.compilers.tarski_grounder",
        "TarskiGrounder",
    ),
    "up_grounder": ("unified_planning.engines.compilers.grounder", "Grounder"),
}

DEFAULT_META_ENGINES = {
    "oversubscription": (
        "unified_planning.engines.oversubscription_planner",
        "OversubscriptionPlanner",
    ),
}

DEFAULT_ENGINES_PREFERENCE_LIST = [
    "fast-downward",
    "fast-downward-opt",
    "pyperplan",
    "enhsp",
    "enhsp-opt",
    "tamer",
    "sequential_plan_validator",
    "sequential_simulator",
    "up_conditional_effects_remover",
    "up_disjunctive_conditions_remover",
    "up_negative_conditions_remover",
    "up_quantifiers_remover",
    "tarski_grounder",
    "up_grounder",
]

DEFAULT_META_ENGINES_PREFERENCE_LIST = ["oversubscription"]


def format_table(header: List[str], rows: List[List[str]]) -> str:
    row_template = "|"
    for i in range(len(header)):
        l = max(len(r[i]) for r in [header] + rows)
        row_template += f" {{:<{str(l)}}} |"
    header_str = row_template.format(*header)
    row_len = len(header_str)
    rows_str = [f'{"-"*row_len}', f"{header_str}", f'{"="*row_len}']
    for row in rows:
        rows_str.append(f"{row_template.format(*row)}")
        rows_str.append(f'{"-"*row_len}')
    return "\n".join(rows_str)


def get_possible_config_locations() -> List[str]:
    """Returns all the possible location of the configuration file."""
    home = os.path.expanduser("~")
    files = []
    stack = inspect.stack()
    for p in PurePath(os.path.abspath(stack[-1].filename)).parents:
        files.append(os.path.join(p, "up.ini"))
        files.append(os.path.join(p, ".up.ini"))
    files.append(os.path.join(home, "up.ini"))
    files.append(os.path.join(home, ".up.ini"))
    files.append(os.path.join(home, ".uprc"))
    return files


class Factory:
    def __init__(self, env: "Environment"):
        self._env = env
        self._engines: Dict[str, Type["up.engines.engine.Engine"]] = {}
        self._engines_info: List[Tuple[str, str, str]] = []
        self._meta_engines: Dict[str, Type["up.engines.meta_engine.MetaEngine"]] = {}
        self._meta_engines_info: List[Tuple[str, str, str]] = []
        self._credit_disclaimer_printed = False
        for name, (module_name, class_name) in DEFAULT_ENGINES.items():
            try:
                self._add_engine(name, module_name, class_name)
            except ImportError:
                pass
        engines = dict(self._engines)
        for name, (module_name, class_name) in DEFAULT_META_ENGINES.items():
            try:
                for engine_name, engine in engines.items():
                    self._add_meta_engine(
                        name, module_name, class_name, engine_name, engine
                    )
            except ImportError:
                pass
        self._preference_list = []
        for name in DEFAULT_ENGINES_PREFERENCE_LIST:
            if name in self._engines:
                self._preference_list.append(name)
        for name in DEFAULT_META_ENGINES_PREFERENCE_LIST:
            for e in self._engines.keys():
                if e.startswith(f"{name}["):
                    self._preference_list.append(e)
        self.configure_from_file()

    # The getstate and setstate method are needed in the Parallel engine.
    # The Parallel engine creates a deep copy of the Factory instance
    # in another process by pickling it.
    # Since local classes are not picklable and engines instantiated from
    # a meta engine are local classes, we need to remove them from the
    # state and then re-create them in the new process.
    def __getstate__(self):
        state = self.__dict__.copy()
        del state["_engines"]
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._engines = {}
        engines_info = list(self._engines_info)
        self._engines_info = []
        for name, module_name, class_name in engines_info:
            self._add_engine(name, module_name, class_name)
        engines = dict(self._engines)
        meta_engines_info = list(self._meta_engines_info)
        self._meta_engines_info = []
        for name, module_name, class_name in meta_engines_info:
            for engine_name, engine in engines.items():
                self._add_meta_engine(
                    name, module_name, class_name, engine_name, engine
                )

    @property
    def engines(self) -> List[str]:
        return list(self._engines.keys())

    def engine(self, name: str) -> Type["up.engines.engine.Engine"]:
        return self._engines[name]

    @property
    def preference_list(self) -> List[str]:
        return self._preference_list

    @preference_list.setter
    def preference_list(self, preference_list: List[str]):
        """Defines the order in which to pick the engines.
        The list is not required to contain all the engines. It is
        possible to define a subsets of the engines, or even just
        one. The impact of this, is that the engine will never be
        selected automatically. Note, however, that the engine can
        still be selected by calling it by name.
        """
        self._preference_list = preference_list

    def add_engine(self, name: str, module_name: str, class_name: str):
        self._add_engine(name, module_name, class_name)
        self._preference_list.append(name)
        engine = self._engines[name]
        for me_name, me in self._meta_engines.items():
            if me.is_compatible_engine(engine):
                n = f"{me_name}[{name}]"
                self._engines[n] = me[engine]
                self._preference_list.append(n)

    def add_meta_engine(self, name: str, module_name: str, class_name: str):
        for engine_name, engine in self._engines.items():
            self._add_meta_engine(name, module_name, class_name, engine_name, engine)
            n = f"{name}[{engine_name}]"
            if n in self.engines:
                self._preference_list.append(n)

    def configure_from_file(self, config_filename: Optional[str] = None):
        """
        Reads a configuration file and configures the factory.

        The following is an example of configuration file:


        [global]
        engine_preference_list: fast-downward fast-downward-opt enhsp enhsp-opt tamer

        [engine <engine-name>]
        module_name: <module-name>
        class_name: <class-name>


        If not given, the configuration is read from the first `up.ini` or `.up.ini` file
        located in any of the parent directories from which this code was called  or,
        otherwise, from one of the following files: `~/up.ini`, `~/.up.ini`, `~/.uprc`.
        """
        config = configparser.ConfigParser()
        if config_filename is None:
            files = get_possible_config_locations()
            config.read(files)
        else:
            config.read([config_filename])

        new_engine_sections = [
            s for s in config.sections() if s.lower().startswith("engine ")
        ]

        for s in new_engine_sections:
            name = s[len("engine ") :]

            module_name = config.get(s, "module_name")
            assert module_name is not None, (
                "Missing 'module_name' value in definition" "of '%s' engine" % name
            )

            class_name = config.get(s, "class_name")
            assert class_name is not None, (
                "Missing 'class_name' value in definition" "of '%s' engine" % name
            )

            self.add_engine(name, module_name, class_name)

        new_meta_engine_sections = [
            s for s in config.sections() if s.lower().startswith("meta-engine ")
        ]

        for s in new_meta_engine_sections:
            name = s[len("meta-engine ") :]

            module_name = config.get(s, "module_name")
            assert module_name is not None, (
                "Missing 'module_name' value in definition of '%s' meta-engine" % name
            )

            class_name = config.get(s, "class_name")
            assert class_name is not None, (
                "Missing 'class_name' value in definition of '%s' meta-engine" % name
            )

            self.add_meta_engine(name, module_name, class_name)

        if "global" in config.sections():
            pref_list = config.get("global", "engine_preference_list")

            if pref_list is not None:
                prefs = [x.strip() for x in pref_list.split() if len(x.strip()) > 0]
                self.preference_list = [e for e in prefs if e in self.engines]

    def _add_engine(self, name: str, module_name: str, class_name: str):
        module = importlib.import_module(module_name)
        EngineImpl = getattr(module, class_name)
        self._engines[name] = EngineImpl
        self._engines_info.append((name, module_name, class_name))

    def _add_meta_engine(
        self,
        name: str,
        module_name: str,
        class_name: str,
        engine_name: str,
        engine: Type["up.engines.engine.Engine"],
    ):
        if name in self._meta_engines:
            EngineImpl = self._meta_engines[name]
        else:
            module = importlib.import_module(module_name)
            EngineImpl = getattr(module, class_name)
            self._meta_engines[name] = EngineImpl
            self._meta_engines_info.append((name, module_name, class_name))
        if EngineImpl.is_compatible_engine(engine):
            self._engines[f"{name}[{engine_name}]"] = EngineImpl[engine]

    def _get_engine_class(
        self,
        engine_kind: str,
        name: Optional[str] = None,
        problem_kind: ProblemKind = ProblemKind(),
        optimality_guarantee: Optional["OptimalityGuarantee"] = None,
        compilation_kind: Optional["CompilationKind"] = None,
        plan_kind: Optional["PlanKind"] = None,
    ) -> Type["up.engines.engine.Engine"]:
        if name is not None:
            if name in self._engines:
                return self._engines[name]
            else:
                raise up.exceptions.UPNoRequestedEngineAvailableException
        problem_features = list(problem_kind.features)
        planners_features = []
        # Make sure that optimality guarantees and compilation kind are mutually exclusive
        assert optimality_guarantee is None or compilation_kind is None
        for name in self._preference_list:
            EngineClass = self._engines[name]
            if getattr(EngineClass, "is_" + engine_kind)():
                assert optimality_guarantee is None or issubclass(
                    EngineClass, OneshotPlannerMixin
                )
                assert compilation_kind is None or issubclass(
                    EngineClass, CompilerMixin
                )
                assert plan_kind is None or issubclass(EngineClass, PlanValidatorMixin)
                if (
                    EngineClass.supports(problem_kind)
                    and (
                        optimality_guarantee is None
                        or cast(OneshotPlannerMixin, EngineClass).satisfies(
                            optimality_guarantee
                        )
                    )
                    and (
                        compilation_kind is None
                        or cast(CompilerMixin, EngineClass).supports_compilation(
                            compilation_kind
                        )
                    )
                    and (
                        plan_kind is None
                        or cast(PlanValidatorMixin, EngineClass).supports_plan(
                            plan_kind
                        )
                    )
                ):
                    return EngineClass
                elif (
                    compilation_kind is None
                    or cast(CompilerMixin, EngineClass).supports_compilation(
                        compilation_kind
                    )
                ) and (
                    plan_kind is None
                    or cast(PlanValidatorMixin, EngineClass).supports_plan(plan_kind)
                ):
                    x = [name] + [
                        str(EngineClass.supports(ProblemKind({f})))
                        for f in problem_features
                    ]
                    if optimality_guarantee is not None:
                        x.append(
                            str(
                                cast(Type[OneshotPlannerMixin], EngineClass).satisfies(
                                    optimality_guarantee
                                )
                            )
                        )
                    planners_features.append(x)
        if len(planners_features) > 0:
            header = ["Engine"] + problem_features
            if optimality_guarantee is not None:
                header.append("OPTIMALITY_GUARANTEE")
            msg = f"No available engine supports all the problem features:\n{format_table(header, planners_features)}"
        elif compilation_kind is not None:
            msg = f"No available engine supports {compilation_kind}"
        elif plan_kind is not None:
            msg = f"No available engine supports {plan_kind}"
        else:
            msg = f"No available {engine_kind} engine"
        raise up.exceptions.UPNoSuitableEngineAvailableException(msg)

    def _print_credits(self, all_credits: List[Optional["up.engines.Credits"]]):
        """
        This function prints the credits of the engine(s) used by an operation mode
        """
        credits: List["up.engines.Credits"] = [c for c in all_credits if c is not None]
        if len(credits) == 0:
            return

        stack = inspect.stack()
        fname = stack[3].filename
        if "unified_planning/shortcuts.py" in fname:
            fname = stack[4].filename
            operation_mode_name = stack[3].function
            line = stack[4].lineno
        else:
            operation_mode_name = stack[2].function
            line = stack[3].lineno

        class PaleWriter(up.AnyBaseClass):
            def __init__(self, stream: IO[str]):
                self._stream = stream

            def write(self, txt: str):
                self._stream.write("\033[96m")
                self._stream.write(txt)
                self._stream.write("\033[0m")

        if self.environment.credits_stream is not None:
            w = PaleWriter(self.environment.credits_stream)

            if not self._credit_disclaimer_printed:
                self._credit_disclaimer_printed = True
                w.write(
                    f"\033[1mNOTE: To disable printing of planning engine credits, add this line to your code: `up.shortcuts.get_env().credits_stream = None`\n"
                )
            w.write("  *** Credits ***\n")
            w.write(
                f"  * In operation mode `{operation_mode_name}` at line {line} of `{fname}`, "
            )
            if len(credits) > 1:
                w.write(
                    "you are using a parallel planning engine with the following components:\n"
                )
            else:
                w.write("you are using the following planning engine:\n")
            for c in credits:
                c.write_credits(w)
            w.write("\n")

    def _get_engine(
        self,
        engine_kind: str,
        name: Optional[str] = None,
        names: Optional[List[str]] = None,
        params: Union[Dict[str, str], List[Dict[str, str]]] = None,
        problem_kind: ProblemKind = ProblemKind(),
        optimality_guarantee: Optional["OptimalityGuarantee"] = None,
        compilation_kind: Optional["CompilationKind"] = None,
        plan_kind: Optional["PlanKind"] = None,
        problem: Optional["up.model.AbstractProblem"] = None,
    ) -> "up.engines.engine.Engine":
        if names is not None:
            assert name is None
            assert problem is None, "Parallel simulation is not supported"
            if params is None:
                params = [{} for i in range(len(names))]
            assert isinstance(params, List) and len(names) == len(params)
            engines = []
            all_credits = []
            for name, param in zip(names, params):
                EngineClass = self._get_engine_class(engine_kind, name)
                all_credits.append(EngineClass.get_credits(**param))
                engines.append((name, param))
            self._print_credits(all_credits)
            p_engine = up.engines.parallel.Parallel(self, engines)
            return p_engine
        else:
            if params is None:
                params = {}
            assert isinstance(params, Dict)
            EngineClass = self._get_engine_class(
                engine_kind,
                name,
                problem_kind,
                optimality_guarantee,
                compilation_kind,
                plan_kind,
            )
            credits = EngineClass.get_credits(**params)
            self._print_credits([credits])
            if problem is None:
                assert engine_kind != "simulator"
                return EngineClass(**params)
            else:
                assert engine_kind == "simulator"
                assert issubclass(EngineClass, up.engines.engine.Engine)
                assert issubclass(
                    EngineClass, up.engines.mixins.simulator.SimulatorMixin
                )
                return EngineClass(problem, **params)

    @property
    def environment(self) -> "Environment":
        """Returns the environment in which this factory is created"""
        return self._env

    def OneshotPlanner(
        self,
        *,
        name: Optional[str] = None,
        names: Optional[List[str]] = None,
        params: Union[Dict[str, str], List[Dict[str, str]]] = None,
        problem_kind: ProblemKind = ProblemKind(),
        optimality_guarantee: Optional[Union["OptimalityGuarantee", str]] = None,
    ) -> "up.engines.engine.Engine":
        """
        Returns a oneshot planner. There are three ways to call this method:
        - using 'name' (the name of a specific planner) and 'params' (planner dependent options).
          e.g. OneshotPlanner(name='tamer', params={'heuristic': 'hadd'})
        - using 'names' (list of specific planners name) and 'params' (list of
          planners dependent options) to get a Parallel engine.
          e.g. OneshotPlanner(names=['tamer', 'tamer'],
                              params=[{'heuristic': 'hadd'}, {'heuristic': 'hmax'}])
        - using 'problem_kind' and 'optimality_guarantee'.
          e.g. OneshotPlanner(problem_kind=problem.kind, optimality_guarantee=SOLVED_OPTIMALLY)
        """
        if isinstance(optimality_guarantee, str):
            optimality_guarantee = OptimalityGuarantee[optimality_guarantee]
        return self._get_engine(
            "oneshot_planner", name, names, params, problem_kind, optimality_guarantee
        )

    def PlanValidator(
        self,
        *,
        name: Optional[str] = None,
        names: Optional[List[str]] = None,
        params: Union[Dict[str, str], List[Dict[str, str]]] = None,
        problem_kind: ProblemKind = ProblemKind(),
        plan_kind: Optional[Union["PlanKind", str]] = None,
    ) -> "up.engines.engine.Engine":
        """
        Returns a plan validator. There are three ways to call this method:
        - using 'name' (the name of a specific plan validator) and 'params'
          (plan validator dependent options).
          e.g. PlanValidator(name='tamer', params={'opt': 'val'})
        - using 'names' (list of specific plan validators name) and 'params' (list of
          plan validators dependent options) to get a Parallel engine.
          e.g. PlanValidator(names=['tamer', 'tamer'],
                             params=[{'opt1': 'val1'}, {'opt2': 'val2'}])
        - using 'problem_kind' and 'plan_kind' parameters.
          e.g. PlanValidator(problem_kind=problem.kind, plan_kind=plan.kind)
        """
        if isinstance(plan_kind, str):
            plan_kind = PlanKind[plan_kind]
        return self._get_engine(
            "plan_validator", name, names, params, problem_kind, plan_kind=plan_kind
        )

    def Compiler(
        self,
        *,
        name: Optional[str] = None,
        params: Union[Dict[str, str], List[Dict[str, str]]] = None,
        problem_kind: ProblemKind = ProblemKind(),
        compilation_kind: Optional[Union["CompilationKind", str]] = None,
    ) -> "up.engines.engine.Engine":
        """
        Returns a Compiler. There are two ways to call this method:
        - using 'name' (the name of a specific grounder) and 'params'
          (grounder dependent options).
          e.g. Compiler(name='tamer', params={'opt': 'val'})
        - using 'problem_kind' and 'compilation_kind' parameters.
          e.g. Compiler(problem_kind=problem.kind, compilation_kind=GROUNDER)
        """
        if isinstance(compilation_kind, str):
            compilation_kind = CompilationKind[compilation_kind]
        return self._get_engine(
            "compiler",
            name,
            None,
            params,
            problem_kind,
            compilation_kind=compilation_kind,
        )

    def Simulator(
        self,
        problem: "up.model.AbstractProblem",
        *,
        name: Optional[str] = None,
        params: Union[Dict[str, str], List[Dict[str, str]]] = None,
    ) -> "up.engines.engine.Engine":
        """
        Returns a Simulator. There are two ways to call this method:
        - using 'problem_kind' through the problem field.
          e.g. Simulator(problem)
        - using 'name' (the name of a specific simulator) and eventually some 'params'
          (simulator dependent options).
          e.g. Simulator(problem, name='sequential_simulator')
        """
        return self._get_engine(
            "simulator", name, None, params, problem.kind, problem=problem
        )

    def print_engines_info(
        self, stream: IO[str] = sys.stdout, full_credits: bool = True
    ):
        stream.write("These are the engines currently available:\n")
        for Engine in self._engines.values():
            credits = Engine.get_credits()
            if credits is not None:
                stream.write("---------------------------------------\n")
                credits.write_credits(stream, full_credits)
                stream.write(
                    f"This engine supports the following features:\n{str(Engine.supported_kind())}\n\n"
                )



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unified_planning
from unified_planning.shortcuts import *
from unified_planning.exceptions import UPProblemDefinitionError
from unified_planning.model import GlobalStartTiming
from unified_planning.model.problem_kind import (
    classical_kind,
    full_classical_kind,
    basic_temporal_kind,
)
from unified_planning.test import TestCase, main
from unified_planning.test import (
    skipIfNoPlanValidatorForProblemKind,
    skipIfNoOneshotPlannerForProblemKind,
)
from unified_planning.test.examples import get_example_problems
from unified_planning.engines.compilers import ConditionalEffectsRemover
from unified_planning.engines import CompilationKind


class TestConditionalEffectsRemover(TestCase):
    def setUp(self):
        TestCase.setUp(self)
        self.problems = get_example_problems()

    @skipIfNoOneshotPlannerForProblemKind(classical_kind)
    @skipIfNoPlanValidatorForProblemKind(full_classical_kind)
    def test_basic_conditional(self):
        problem = self.problems["basic_conditional"].problem
        with Compiler(
            problem_kind=problem.kind,
            compilation_kind=CompilationKind.CONDITIONAL_EFFECTS_REMOVING,
        ) as cer:
            res = cer.compile(problem, CompilationKind.CONDITIONAL_EFFECTS_REMOVING)
        unconditional_problem = res.problem

        self.assertTrue(problem.kind.has_conditional_effects())
        self.assertFalse(unconditional_problem.kind.has_conditional_effects())
        for a in unconditional_problem.actions:
            self.assertFalse(a.is_conditional())

        with OneshotPlanner(problem_kind=unconditional_problem.kind) as planner:
            self.assertNotEqual(planner, None)
            uncond_plan = planner.solve(unconditional_problem).plan
            new_plan = uncond_plan.replace_action_instances(
                res.map_back_action_instance
            )
            with PlanValidator(
                problem_kind=problem.kind, plan_kind=new_plan.kind
            ) as pv:
                self.assertTrue(pv.validate(problem, new_plan))

    @skipIfNoOneshotPlannerForProblemKind(classical_kind)
    @skipIfNoPlanValidatorForProblemKind(full_classical_kind)
    def test_complex_conditional(self):
        problem = self.problems["complex_conditional"].problem
        with Compiler(name="up_conditional_effects_remover") as cer:
            res = cer.compile(problem, CompilationKind.CONDITIONAL_EFFECTS_REMOVING)
        unconditional_problem = res.problem
        res_2 = cer.compile(problem, CompilationKind.CONDITIONAL_EFFECTS_REMOVING)
        unconditional_problem_2 = res_2.problem
        self.assertEqual(unconditional_problem, unconditional_problem_2)

        with OneshotPlanner(problem_kind=unconditional_problem.kind) as planner:
            self.assertNotEqual(planner, None)
            uncond_plan = planner.solve(unconditional_problem).plan
            new_plan = uncond_plan.replace_action_instances(
                res.map_back_action_instance
            )
            with PlanValidator(
                problem_kind=problem.kind, plan_kind=new_plan.kind
            ) as pv:
                self.assertTrue(pv.validate(problem, new_plan))

    @skipIfNoOneshotPlannerForProblemKind(classical_kind.union(basic_temporal_kind))
    def test_temporal_conditional(self):
        problem = self.problems["temporal_conditional"].problem

        with Compiler(name="up_conditional_effects_remover") as cer:
            res = cer.compile(problem, CompilationKind.CONDITIONAL_EFFECTS_REMOVING)
        unconditional_problem = res.problem

        with OneshotPlanner(problem_kind=unconditional_problem.kind) as planner:
            self.assertNotEqual(planner, None)
            uncond_plan = planner.solve(unconditional_problem).plan
            new_plan = uncond_plan.replace_action_instances(
                res.map_back_action_instance
            )
            for (s, a, d), (s_1, a_1, d_1) in zip(
                new_plan.timed_actions, uncond_plan.timed_actions
            ):
                self.assertEqual(s, s_1)
                self.assertEqual(d, d_1)
                self.assertIn(a.action, problem.actions)

    def test_ad_hoc_1(self):
        ct = GlobalStartTiming(2)
        x = unified_planning.model.Fluent("x")
        y = unified_planning.model.Fluent("y")
        problem = unified_planning.model.Problem("ad_hoc_1")
        problem.add_fluent(x, default_initial_value=True)
        problem.add_fluent(y, default_initial_value=True)
        problem.add_timed_effect(ct, y, Not(x), x)
        problem.add_goal(Not(y))
        cer = ConditionalEffectsRemover()
        res = cer.compile(problem, CompilationKind.CONDITIONAL_EFFECTS_REMOVING)
        uncond_problem = res.problem
        eff = uncond_problem.timed_effects[ct][0]
        self.assertFalse(eff.is_conditional())
        self.assertEqual(FluentExp(y), eff.fluent)
        self.assertEqual(And(Not(x), y), eff.value)

    def test_mockup_1(self):
        ct = GlobalStartTiming(2)
        x = unified_planning.model.Fluent("x", IntType())
        y = unified_planning.model.Fluent("y")
        problem = unified_planning.model.Problem("mockup_1")
        problem.add_fluent(x, default_initial_value=0)
        problem.add_fluent(y, default_initial_value=True)
        problem.add_timed_effect(ct, y, False)
        problem.add_timed_effect(ct, x, 5, y)
        cer = ConditionalEffectsRemover()
        with self.assertRaises(UPProblemDefinitionError) as e:
            cer.compile(problem, CompilationKind.CONDITIONAL_EFFECTS_REMOVING)
        self.assertIn(
            "The condition of effect: if y then x := 5\ncould not be removed without changing the problem.",
            str(e.exception),
        )



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License

import os
import tempfile
from typing import cast
import pytest
import unified_planning
from unified_planning.shortcuts import *
from unified_planning.test import TestCase, main
from unified_planning.io import PDDLWriter, PDDLReader
from unified_planning.test.examples import get_example_problems
from unified_planning.model.types import _UserType


FILE_PATH = os.path.dirname(os.path.abspath(__file__))
PDDL_DOMAINS_PATH = os.path.join(FILE_PATH, "pddl")


class TestPddlIO(TestCase):
    def setUp(self):
        TestCase.setUp(self)
        self.problems = get_example_problems()

    def test_basic_writer(self):
        problem = self.problems["basic"].problem

        w = PDDLWriter(problem)

        pddl_domain = w.get_domain()
        self.assertIn("(:requirements :strips :negative-preconditions)", pddl_domain)
        self.assertIn("(:predicates (x))", pddl_domain)
        self.assertIn("(:action a", pddl_domain)
        self.assertIn(":parameters ()", pddl_domain)
        self.assertIn(":precondition (and (not (x)))", pddl_domain)
        self.assertIn(":effect (and (x))", pddl_domain)

        pddl_problem = w.get_problem()
        self.assertIn("(:domain basic-domain)", pddl_problem)
        self.assertIn("(:init)", pddl_problem)
        self.assertIn("(:goal (and (x)))", pddl_problem)

    def test_basic_conditional_writer(self):
        problem = self.problems["basic_conditional"].problem

        self.assertTrue(problem.action("a_x").is_conditional())
        self.assertFalse(problem.action("a_y").is_conditional())

        w = PDDLWriter(problem)

        pddl_domain = w.get_domain()
        self.assertIn(
            "(:requirements :strips :negative-preconditions :conditional-effects)",
            pddl_domain,
        )
        self.assertIn("(:predicates (x) (y))", pddl_domain)
        self.assertIn("(:action a_x", pddl_domain)
        self.assertIn(":parameters ()", pddl_domain)
        self.assertIn(":precondition (and (not (x)))", pddl_domain)
        self.assertIn(":effect (and (when (y) (x)))", pddl_domain)
        self.assertIn("(:action a_y", pddl_domain)
        self.assertIn(":parameters ()", pddl_domain)
        self.assertIn(":precondition (and (not (y)))", pddl_domain)
        self.assertIn(":effect (and (y))", pddl_domain)

        pddl_problem = w.get_problem()
        self.assertIn("(:domain basic_conditional-domain)", pddl_problem)
        self.assertIn("(:init)", pddl_problem)
        self.assertIn("(:goal (and (x)))", pddl_problem)

    def test_basic_exists_writer(self):
        problem = self.problems["basic_exists"].problem

        w = PDDLWriter(problem)

        pddl_domain = w.get_domain()
        self.assertIn(
            "(:requirements :strips :typing :existential-preconditions)", pddl_domain
        )
        self.assertIn("(:predicates (x) (y ?semaphore - semaphore))", pddl_domain)
        self.assertIn("(:action a", pddl_domain)
        self.assertIn(":parameters ()", pddl_domain)
        self.assertIn(
            ":precondition (and (exists (?s - semaphore)\n (y ?s)))", pddl_domain
        )
        self.assertIn(":effect (and (x))", pddl_domain)

        pddl_problem = w.get_problem()
        self.assertIn("(:domain basic_exists-domain)", pddl_problem)
        self.assertIn("(:objects\n   o1 o2 - semaphore\n )", pddl_problem)
        self.assertIn("(:init (y o1))", pddl_problem)
        self.assertIn("(:goal (and (x)))", pddl_problem)

    def test_robot_writer(self):
        problem = self.problems["robot"].problem

        w = PDDLWriter(problem)

        pddl_domain = w.get_domain()
        self.assertIn(
            "(:requirements :strips :typing :negative-preconditions :equality :numeric-fluents)",
            pddl_domain,
        )
        self.assertIn("(:types location)", pddl_domain)
        self.assertIn("(:predicates (robot_at ?position - location))", pddl_domain)
        self.assertIn("(:functions (battery_charge))", pddl_domain)
        self.assertIn("(:action move", pddl_domain)
        self.assertIn(":parameters ( ?l_from - location ?l_to - location)", pddl_domain)
        self.assertIn(
            ":precondition (and (<= 10 (battery_charge)) (not (= ?l_from ?l_to)) (robot_at ?l_from) (not (robot_at ?l_to)))",
            pddl_domain,
        )
        self.assertIn(
            ":effect (and (not (robot_at ?l_from)) (robot_at ?l_to) (assign (battery_charge) (- (battery_charge) 10)))",
            pddl_domain,
        )

        pddl_problem = w.get_problem()
        self.assertIn("(:domain robot-domain)", pddl_problem)
        self.assertIn("(:objects", pddl_problem)
        self.assertIn("l1 l2 - location", pddl_problem)
        self.assertIn("(:init (robot_at l1) (= (battery_charge) 100))", pddl_problem)
        self.assertIn("(:goal (and (robot_at l2)))", pddl_problem)

    def test_robot_decrease_writer(self):
        problem = self.problems["robot_decrease"].problem

        w = PDDLWriter(problem)

        pddl_domain = w.get_domain()
        self.assertIn(
            "(:requirements :strips :typing :negative-preconditions :equality :numeric-fluents)",
            pddl_domain,
        )
        self.assertIn("(:types location)", pddl_domain)
        self.assertIn("(:predicates (robot_at ?position - location))", pddl_domain)
        self.assertIn("(:functions (battery_charge))", pddl_domain)
        self.assertIn("(:action move", pddl_domain)
        self.assertIn(":parameters ( ?l_from - location ?l_to - location)", pddl_domain)
        self.assertIn(
            ":precondition (and (<= 10 (battery_charge)) (not (= ?l_from ?l_to)) (robot_at ?l_from) (not (robot_at ?l_to)))",
            pddl_domain,
        )
        self.assertIn(
            ":effect (and (not (robot_at ?l_from)) (robot_at ?l_to) (decrease (battery_charge) 10))",
            pddl_domain,
        )

        pddl_problem = w.get_problem()
        self.assertIn("(:domain robot_decrease-domain)", pddl_problem)
        self.assertIn("(:objects", pddl_problem)
        self.assertIn("l1 l2 - location", pddl_problem)
        self.assertIn("(:init (robot_at l1) (= (battery_charge) 100))", pddl_problem)
        self.assertIn("(:goal (and (robot_at l2)))", pddl_problem)

    def test_robot_loader_writer(self):
        problem = self.problems["robot_loader"].problem

        w = PDDLWriter(problem)

        pddl_domain = w.get_domain()
        self.assertIn(
            "(:requirements :strips :typing :negative-preconditions :equality)",
            pddl_domain,
        )
        self.assertIn("(:types location)", pddl_domain)
        self.assertIn(
            "(:predicates (robot_at ?position - location) (cargo_at ?position - location) (cargo_mounted))",
            pddl_domain,
        )
        self.assertIn("(:action move", pddl_domain)
        self.assertIn(":parameters ( ?l_from - location ?l_to - location)", pddl_domain)
        self.assertIn(
            ":precondition (and (not (= ?l_from ?l_to)) (robot_at ?l_from) (not (robot_at ?l_to)))",
            pddl_domain,
        )
        self.assertIn(
            ":effect (and (not (robot_at ?l_from)) (robot_at ?l_to))", pddl_domain
        )
        self.assertIn("(:action load", pddl_domain)
        self.assertIn(":parameters ( ?loc - location)", pddl_domain)
        self.assertIn(
            ":precondition (and (cargo_at ?loc) (robot_at ?loc) (not (cargo_mounted)))",
            pddl_domain,
        )
        self.assertIn(
            ":effect (and (not (cargo_at ?loc)) (cargo_mounted))", pddl_domain
        )
        self.assertIn("(:action unload", pddl_domain)
        self.assertIn(":parameters ( ?loc - location)", pddl_domain)
        self.assertIn(
            ":precondition (and (not (cargo_at ?loc)) (robot_at ?loc) (cargo_mounted))",
            pddl_domain,
        )
        self.assertIn(
            ":effect (and (cargo_at ?loc) (not (cargo_mounted)))", pddl_domain
        )

        pddl_problem = w.get_problem()
        self.assertIn("(:domain robot_loader-domain)", pddl_problem)
        self.assertIn("(:objects", pddl_problem)
        self.assertIn("l1 l2 - location", pddl_problem)
        self.assertIn("(:init (robot_at l1) (cargo_at l2))", pddl_problem)
        self.assertIn("(:goal (and (cargo_at l1)))", pddl_problem)

    def test_robot_loader_adv_writer(self):
        problem = self.problems["robot_loader_adv"].problem

        w = PDDLWriter(problem)

        pddl_domain = w.get_domain()
        self.assertIn(
            "(:requirements :strips :typing :negative-preconditions :equality)",
            pddl_domain,
        )
        self.assertIn("(:types robot location container)", pddl_domain)
        self.assertIn(
            "(:predicates (robot_at ?robot - robot ?position - location) (cargo_at ?cargo - container ?position - location) (cargo_mounted ?cargo - container ?robot - robot))",
            pddl_domain,
        )
        self.assertIn("(:action move", pddl_domain)
        self.assertIn(
            ":parameters ( ?l_from - location ?l_to - location ?r - robot)", pddl_domain
        )
        self.assertIn(
            ":precondition (and (not (= ?l_from ?l_to)) (robot_at ?r ?l_from) (not (robot_at ?r ?l_to)))",
            pddl_domain,
        )
        self.assertIn(
            ":effect (and (not (robot_at ?r ?l_from)) (robot_at ?r ?l_to))", pddl_domain
        )
        self.assertIn("(:action load", pddl_domain)
        self.assertIn(
            ":parameters ( ?loc - location ?r - robot ?c - container)", pddl_domain
        )
        self.assertIn(
            ":precondition (and (cargo_at ?c ?loc) (robot_at ?r ?loc) (not (cargo_mounted ?c ?r)))",
            pddl_domain,
        )
        self.assertIn(
            ":effect (and (not (cargo_at ?c ?loc)) (cargo_mounted ?c ?r))", pddl_domain
        )
        self.assertIn("(:action unload", pddl_domain)
        self.assertIn(
            ":parameters ( ?loc - location ?r - robot ?c - container)", pddl_domain
        )
        self.assertIn(
            ":precondition (and (not (cargo_at ?c ?loc)) (robot_at ?r ?loc) (cargo_mounted ?c ?r))",
            pddl_domain,
        )
        self.assertIn(
            ":effect (and (cargo_at ?c ?loc) (not (cargo_mounted ?c ?r)))", pddl_domain
        )

        pddl_problem = w.get_problem()
        self.assertIn("(:domain robot_loader_adv-domain)", pddl_problem)
        self.assertIn("(:objects", pddl_problem)
        self.assertIn("r1 - robot", pddl_problem)
        self.assertIn("l1 l2 l3 - location", pddl_problem)
        self.assertIn("c1 - container", pddl_problem)
        self.assertIn("(:init (robot_at r1 l1) (cargo_at c1 l2))", pddl_problem)
        self.assertIn("(:goal (and (cargo_at c1 l3) (robot_at r1 l1)))", pddl_problem)

    def test_matchcellar_writer(self):
        problem = self.problems["matchcellar"].problem

        w = PDDLWriter(problem)

        pddl_domain = w.get_domain()
        self.assertIn("(define (domain matchcellar-domain)", pddl_domain)
        self.assertIn(
            "(:requirements :strips :typing :negative-preconditions :durative-actions)",
            pddl_domain,
        )
        self.assertIn("(:types match fuse)", pddl_domain)
        self.assertIn(
            "(:predicates (handfree) (light) (match_used ?match - match) (fuse_mended ?fuse - fuse))",
            pddl_domain,
        )
        self.assertIn("(:durative-action light_match", pddl_domain)
        self.assertIn(":parameters ( ?m - match)", pddl_domain)
        self.assertIn(":duration (= ?duration 6)", pddl_domain)
        self.assertIn(":condition (and (at start (not (match_used ?m))))", pddl_domain)
        self.assertIn(
            ":effect (and (at start (match_used ?m)) (at start (light)) (at end (not (light)))))",
            pddl_domain,
        )
        self.assertIn("(:durative-action mend_fuse", pddl_domain)
        self.assertIn(":parameters ( ?f - fuse)", pddl_domain)
        self.assertIn(":duration (= ?duration 5)", pddl_domain)
        self.assertIn(
            ":condition (and (at start (handfree))(at start (light))(over all (light))(at end (light)))",
            pddl_domain,
        )
        self.assertIn(
            ":effect (and (at start (not (handfree))) (at end (fuse_mended ?f)) (at end (handfree))))",
            pddl_domain,
        )

    def test_depot_reader(self):
        reader = PDDLReader()

        domain_filename = os.path.join(PDDL_DOMAINS_PATH, "depot", "domain.pddl")
        problem_filename = os.path.join(PDDL_DOMAINS_PATH, "depot", "problem.pddl")
        problem = reader.parse_problem(domain_filename, problem_filename)

        self.assertTrue(problem is not None)
        self.assertEqual(len(problem.fluents), 15)
        self.assertEqual(len(problem.actions), 5)
        self.assertEqual(len(list(problem.objects(problem.user_type("object")))), 13)

    def test_counters_reader(self):
        reader = PDDLReader()

        domain_filename = os.path.join(PDDL_DOMAINS_PATH, "counters", "domain.pddl")
        problem_filename = os.path.join(PDDL_DOMAINS_PATH, "counters", "problem.pddl")
        problem = reader.parse_problem(domain_filename, problem_filename)

        self.assertTrue(problem is not None)
        self.assertEqual(len(problem.fluents), 2)
        self.assertEqual(len(problem.actions), 2)
        self.assertEqual(len(list(problem.objects(problem.user_type("counter")))), 4)

    def test_sailing_reader(self):
        reader = PDDLReader()

        domain_filename = os.path.join(PDDL_DOMAINS_PATH, "sailing", "domain.pddl")
        problem_filename = os.path.join(PDDL_DOMAINS_PATH, "sailing", "problem.pddl")
        problem = reader.parse_problem(domain_filename, problem_filename)

        self.assertTrue(problem is not None)
        self.assertEqual(len(problem.fluents), 4)
        self.assertEqual(len(problem.actions), 8)
        self.assertEqual(len(list(problem.objects(problem.user_type("boat")))), 2)
        self.assertEqual(len(list(problem.objects(problem.user_type("person")))), 2)

    def test_matchcellar_reader(self):
        reader = PDDLReader()

        domain_filename = os.path.join(PDDL_DOMAINS_PATH, "matchcellar", "domain.pddl")
        problem_filename = os.path.join(
            PDDL_DOMAINS_PATH, "matchcellar", "problem.pddl"
        )
        problem = reader.parse_problem(domain_filename, problem_filename)

        self.assertTrue(problem is not None)
        self.assertEqual(len(problem.fluents), 4)
        self.assertEqual(len(problem.actions), 2)
        self.assertEqual(len(list(problem.objects(problem.user_type("match")))), 3)
        self.assertEqual(len(list(problem.objects(problem.user_type("fuse")))), 3)

    def test_htn_transport_reader(self):
        reader = PDDLReader()

        domain_filename = os.path.join(
            PDDL_DOMAINS_PATH, "htn-transport", "domain.hddl"
        )
        problem_filename = os.path.join(
            PDDL_DOMAINS_PATH, "htn-transport", "problem.hddl"
        )
        problem = reader.parse_problem(domain_filename, problem_filename)

        assert isinstance(problem, up.model.htn.HierarchicalProblem)
        self.assertEqual(5, len(problem.fluents))
        self.assertEqual(4, len(problem.actions))
        self.assertEqual(
            ["deliver", "get-to", "load", "unload"],
            [task.name for task in problem.tasks],
        )
        self.assertEqual(
            [
                "m-deliver",
                "m-unload",
                "m-load",
                "m-drive-to",
                "m-drive-to-via",
                "m-i-am-there",
            ],
            [method.name for method in problem.methods],
        )
        self.assertEqual(1, len(problem.method("m-drive-to").subtasks))
        self.assertEqual(2, len(problem.method("m-drive-to-via").subtasks))
        self.assertEqual(2, len(problem.task_network.subtasks))

    def test_examples_io(self):
        for example in self.problems.values():
            problem = example.problem
            kind = problem.kind
            if (
                kind.has_intermediate_conditions_and_effects()
                or kind.has_object_fluents()
                or kind.has_oversubscription()
            ):
                continue
            with tempfile.TemporaryDirectory() as tempdir:
                domain_filename = os.path.join(tempdir, "domain.pddl")
                problem_filename = os.path.join(tempdir, "problem.pddl")

                w = PDDLWriter(problem)
                w.write_domain(domain_filename)
                w.write_problem(problem_filename)

                reader = PDDLReader()
                parsed_problem = reader.parse_problem(domain_filename, problem_filename)

                self.assertEqual(len(problem.fluents), len(parsed_problem.fluents))
                self.assertTrue(
                    _have_same_user_types_considering_renamings(
                        problem, parsed_problem, w.get_item_named
                    )
                )
                self.assertEqual(len(problem.actions), len(parsed_problem.actions))
                for a in problem.actions:
                    parsed_a = parsed_problem.action(w.get_pddl_name(a))
                    self.assertEqual(a, w.get_item_named(parsed_a.name))
                    for param, parsed_param in zip(a.parameters, parsed_a.parameters):
                        self.assertEqual(
                            param.type, w.get_item_named(parsed_param.type.name)
                        )
                    if isinstance(a, unified_planning.model.InstantaneousAction):
                        self.assertEqual(len(a.effects), len(parsed_a.effects))
                    elif isinstance(a, unified_planning.model.DurativeAction):
                        self.assertEqual(str(a.duration), str(parsed_a.duration))
                        for t, e in a.effects.items():
                            self.assertEqual(len(e), len(parsed_a.effects[t]))

    def test_basic_with_object_constant(self):
        problem = self.problems["basic_with_object_constant"].problem

        w = PDDLWriter(problem)

        pddl_domain = w.get_domain()
        pddl_problem = w.get_problem()

        expected_domain = """(define (domain basic_with_object_constant-domain)
 (:requirements :strips :typing :negative-preconditions)
 (:types location)
 (:constants
   l1 - location
 )
 (:predicates (is_at ?loc - location))
 (:action move
  :parameters ( ?l_from - location ?l_to - location)
  :precondition (and (is_at ?l_from) (not (is_at ?l_to)))
  :effect (and (not (is_at ?l_from)) (is_at ?l_to)))
 (:action move_to_l1
  :parameters ( ?l_from - location)
  :precondition (and (is_at ?l_from) (not (is_at l1)))
  :effect (and (not (is_at ?l_from)) (is_at l1)))
)
"""
        expected_problem = """(define (problem basic_with_object_constant-problem)
 (:domain basic_with_object_constant-domain)
 (:objects
   l2 - location
 )
 (:init (is_at l1))
 (:goal (and (is_at l2)))
)
"""
        self.assertEqual(pddl_domain, expected_domain)
        self.assertEqual(pddl_problem, expected_problem)

    def test_rationals(self):
        problem = self.problems["robot_decrease"].problem.clone()

        # Check perfect conversion
        battery = problem.fluent("battery_charge")
        problem.set_initial_value(battery, Fraction(5, 2))
        w = PDDLWriter(problem)
        pddl_txt = w.get_problem()
        self.assertNotIn("5/2", pddl_txt)
        self.assertIn("2.5", pddl_txt)

        # Check imperfect conversion
        with pytest.warns(UserWarning, match="cannot exactly represent") as warns:
            battery = problem.fluent("battery_charge")
            problem.set_initial_value(battery, Fraction(10, 3))
            w = PDDLWriter(problem)
            pddl_txt = w.get_problem()
            self.assertNotIn("10/3", pddl_txt)
            self.assertIn("3.333333333", pddl_txt)

    def test_ad_hoc_1(self):
        when = UserType("when")
        fl = Fluent("4ction")
        obj_1 = Object("obj_1", when)
        obj_2 = Object("OBJ_1", when)
        act = InstantaneousAction("forall", AND=when)
        fluent = act.parameter("AND")
        act.add_effect(fl, True, Equals(fluent, obj_1))
        problem = Problem("ad_hoc")
        problem.add_fluent(fl)
        problem.add_action(act)
        problem.add_object(obj_1)
        problem.add_object(obj_2)
        problem.set_initial_value(fl, False)
        w = PDDLWriter(problem)
        pddl_domain = w.get_domain()
        pddl_problem = w.get_problem()
        expected_domain = """(define (domain ad_hoc-domain)
 (:requirements :strips :typing :equality :conditional-effects)
 (:types when_)
 (:constants
   obj_1 - when_
 )
 (:predicates (f_4ction))
 (:action forall_
  :parameters ( ?and_ - when_)
  :effect (and (when (= ?and_ obj_1) (f_4ction))))
)
"""
        expected_problem = """(define (problem ad_hoc-problem)
 (:domain ad_hoc-domain)
 (:objects
   obj_1_0 - when_
 )
 (:init)
 (:goal (and ))
)
"""
        self.assertEqual(pddl_domain, expected_domain)
        self.assertEqual(pddl_problem, expected_problem)


def _have_same_user_types_considering_renamings(
    original_problem: unified_planning.model.Problem,
    tested_problem: unified_planning.model.Problem,
    get_item_named,
) -> bool:
    for tested_type in tested_problem.user_types:
        if (
            get_item_named(cast(_UserType, tested_type).name)
            not in original_problem.user_types
        ):
            return False
    return True



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""This module defines all the types."""

import unified_planning
from fractions import Fraction
from typing import Iterator, Optional, Dict, Tuple, cast
from unified_planning.exceptions import UPProblemDefinitionError, UPTypeError


class Type:
    """Basic class for representing a type."""

    def is_bool_type(self) -> bool:
        """Returns true iff is boolean type."""
        return False

    def is_user_type(self) -> bool:
        """Returns true iff is a user type."""
        return False

    def is_real_type(self) -> bool:
        """Returns true iff is real type."""
        return False

    def is_int_type(self) -> bool:
        """Returns true iff is integer type."""
        return False

    def is_time_type(self) -> bool:
        """Returns true iff is integer type."""
        return False

    def is_compatible(self, t_right: "Type") -> bool:
        """Returns True if the type t_right can be assigned to a fluent that which type is self."""
        return is_compatible_type(self, t_right)


class _BoolType(Type):
    """Represents the boolean type."""

    def __repr__(self) -> str:
        return "bool"

    def is_bool_type(self) -> bool:
        """Returns true iff is boolean type."""
        return True


class _TimeType(Type):
    """Represent the type for an absolute Time"""

    def __repr__(self) -> str:
        return "time"

    def is_time_type(self) -> bool:
        """Returns true iff is boolean type."""
        return True


class _UserType(Type):
    """Represents the user type."""

    def __init__(self, name: str, father: Optional[Type] = None):
        Type.__init__(self)
        self._name = name
        if father is not None and (not father.is_user_type()):
            raise UPTypeError("father field of a UserType must be a UserType.")
        self._father = father

    def __repr__(self) -> str:
        return (
            self._name
            if self._father is None
            else f"{self._name} - {cast(_UserType, self._father).name}"
        )

    @property
    def name(self) -> str:
        """Returns the type name."""
        return self._name

    @property
    def father(self) -> Optional[Type]:
        """Returns the type s father."""
        return self._father

    @property
    def ancestors(self) -> Iterator[Type]:
        """Returns all the ancestors of the given UserType, including itself."""
        type: Optional[Type] = self
        while type is not None:
            yield type
            type = cast(_UserType, type).father

    def is_user_type(self) -> bool:
        """Returns true iff is a user type."""
        return True

    def is_subtype(self, t: Type) -> bool:
        """Returns true iff is a subtype of the given type."""
        assert t.is_user_type()
        p: Optional[Type] = self
        while p is not None:
            if p == t:
                return True
            p = cast(_UserType, p).father
        return False


class _IntType(Type):
    def __init__(self, lower_bound: int = None, upper_bound: int = None):
        Type.__init__(self)
        self._lower_bound = lower_bound
        self._upper_bound = upper_bound

    def __repr__(self) -> str:
        b = []
        if (not self.lower_bound is None) or (not self.upper_bound is None):
            b.append("[")
            b.append("-inf" if self.lower_bound is None else str(self.lower_bound))
            b.append(", ")
            b.append("inf" if self.upper_bound is None else str(self.upper_bound))
            b.append("]")
        return "integer" + "".join(b)

    @property
    def lower_bound(self) -> Optional[int]:
        return self._lower_bound

    @property
    def upper_bound(self) -> Optional[int]:
        return self._upper_bound

    def is_int_type(self) -> bool:
        return True


class _RealType(Type):
    def __init__(self, lower_bound: Fraction = None, upper_bound: Fraction = None):
        Type.__init__(self)
        self._lower_bound = lower_bound
        self._upper_bound = upper_bound

    def __repr__(self) -> str:
        b = []
        if (not self.lower_bound is None) or (not self.upper_bound is None):
            b.append("[")
            b.append("-inf" if self.lower_bound is None else str(self.lower_bound))
            b.append(", ")
            b.append("inf" if self.upper_bound is None else str(self.upper_bound))
            b.append("]")
        return "real" + "".join(b)

    @property
    def lower_bound(self) -> Optional[Fraction]:
        return self._lower_bound

    @property
    def upper_bound(self) -> Optional[Fraction]:
        return self._upper_bound

    def is_real_type(self) -> bool:
        return True


BOOL = _BoolType()
TIME = _TimeType()


class TypeManager:
    def __init__(self):
        self._bool = BOOL
        self._ints: Dict[Tuple[Optional[int], Optional[int]], Type] = {}
        self._reals: Dict[Tuple[Optional[Fraction], Optional[Fraction]], Type] = {}
        self._user_types: Dict[Tuple[str, Optional[Type]], Type] = {}

    def has_type(self, type: Type) -> bool:
        if type.is_bool_type():
            return type == self._bool
        elif type.is_int_type():
            assert isinstance(type, _IntType)
            return self._ints.get((type.lower_bound, type.upper_bound), None) == type
        elif type.is_real_type():
            assert isinstance(type, _RealType)
            return self._reals.get((type.lower_bound, type.upper_bound), None) == type
        elif type.is_time_type():
            return type == TIME
        elif type.is_user_type():
            assert isinstance(type, _UserType)
            return self._user_types.get((type.name, type.father), None) == type
        else:
            raise NotImplementedError

    def BoolType(self) -> Type:
        return self._bool

    def IntType(self, lower_bound: int = None, upper_bound: int = None) -> Type:
        k = (lower_bound, upper_bound)
        if k in self._ints:
            return self._ints[k]
        else:
            it = _IntType(lower_bound, upper_bound)
            self._ints[k] = it
            return it

    def RealType(
        self, lower_bound: Fraction = None, upper_bound: Fraction = None
    ) -> Type:
        k = (lower_bound, upper_bound)
        if k in self._reals:
            return self._reals[k]
        else:
            rt = _RealType(lower_bound, upper_bound)
            self._reals[k] = rt
            return rt

    def UserType(self, name: str, father: Optional[Type] = None) -> Type:
        if (name, father) in self._user_types:
            return self._user_types[(name, father)]
        else:
            if father is not None:
                assert isinstance(father, _UserType)
                if any(
                    cast(_UserType, ancestor).name == name
                    for ancestor in father.ancestors
                ):
                    raise UPTypeError(
                        f"The name: {name} is already used. A UserType and one of his ancestors can not share the name."
                    )
            ut = _UserType(name, father)
            self._user_types[(name, father)] = ut
            return ut


def domain_size(
    objects_set: "unified_planning.model.mixins.ObjectsSetMixin",
    typename: "unified_planning.model.types.Type",
) -> int:
    """Returns the domain size of the given type."""
    if typename.is_bool_type():
        return 2
    elif typename.is_user_type():
        return len(list(objects_set.objects(typename)))
    elif typename.is_int_type():
        typename = cast(_IntType, typename)
        lb = typename.lower_bound
        ub = typename.upper_bound
        if lb is None or ub is None:
            raise UPProblemDefinitionError("Parameter not groundable!")
        return ub - lb
    else:
        raise UPProblemDefinitionError("Parameter not groundable!")


def domain_item(
    objects_set: "unified_planning.model.mixins.ObjectsSetMixin",
    typename: "unified_planning.model.types.Type",
    idx: int,
) -> "unified_planning.model.fnode.FNode":
    """Returns the ith domain item of the given type."""
    if typename.is_bool_type():
        return objects_set.env.expression_manager.Bool(idx == 0)
    elif typename.is_user_type():
        return objects_set.env.expression_manager.ObjectExp(
            list(objects_set.objects(typename))[idx]
        )
    elif typename.is_int_type():
        typename = cast(_IntType, typename)
        lb = typename.lower_bound
        ub = typename.upper_bound
        if lb is None or ub is None:
            raise UPProblemDefinitionError("Parameter not groundable!")
        return objects_set.env.expression_manager.Int(lb + idx)
    else:
        raise UPProblemDefinitionError("Parameter not groundable!")


def is_compatible_type(
    t_left: "Type",
    t_right: "Type",
) -> bool:
    """Returns True if the type t_right can be assigned to a typed up object that has type t_left.
    :param t_left: the target type for the assignment.
    :param t_right: the type of the element that wants to be assigned to the element of type t_left.
    :return: True if the element of type t_left can be assigned to the element of type t_right; False otherwise."""
    if t_left == t_right:
        return True
    if t_left.is_user_type() and t_right.is_user_type():
        assert isinstance(t_left, _UserType) and isinstance(t_right, _UserType)
        return t_right in t_left.ancestors
    if not (
        (t_left.is_int_type() and t_right.is_int_type())
        or (t_left.is_real_type() and t_right.is_real_type())
        or (t_left.is_real_type() and t_right.is_int_type())
    ):
        return False
    assert isinstance(t_left, _IntType) or isinstance(t_left, _RealType)
    assert isinstance(t_right, _IntType) or isinstance(t_right, _RealType)
    left_lower = -float("inf") if t_left.lower_bound is None else t_left.lower_bound
    left_upper = float("inf") if t_left.upper_bound is None else t_left.upper_bound
    right_lower = -float("inf") if t_right.lower_bound is None else t_right.lower_bound
    right_upper = float("inf") if t_right.upper_bound is None else t_right.upper_bound
    if right_upper < left_lower or right_lower > left_upper:
        return False
    else:
        return True



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unified_planning.model.walkers as walkers
from unified_planning.model.fnode import FNode
from unified_planning.model.operators import OperatorKind
from typing import List, Set


class FreeVarsExtractor(walkers.dag.DagWalker):
    """This expression walker returns all the free vars of a given expression."""

    def __init__(self):
        walkers.dag.DagWalker.__init__(self)

    def get(self, expression: FNode) -> Set[FNode]:
        """Returns all the free vars of the given expression."""
        return self.walk(expression)

    @walkers.handles(OperatorKind)
    def walk_all_types(self, expression: FNode, args: List[Set[FNode]]) -> Set[FNode]:
        res = set(x for y in args for x in y)
        if expression.is_fluent_exp():
            res = res | {expression}
        return res



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unified_planning

from unified_planning.shortcuts import *
from unified_planning.test import (
    TestCase,
    skipIfNoOneshotPlannerForProblemKind,
    skipIfNoPlanValidatorForProblemKind,
    skipIfEngineNotAvailable,
)
from unified_planning.test.examples import get_example_problems
from unified_planning.engines import CompilationKind
from unified_planning.model.problem_kind import (
    basic_classical_kind,
    full_classical_kind,
    hierarchical_kind,
)


class TestTarskiGrounder(TestCase):
    def setUp(self):
        TestCase.setUp(self)
        self.problems = get_example_problems()

    @skipIfNoOneshotPlannerForProblemKind(full_classical_kind)
    @skipIfNoPlanValidatorForProblemKind(full_classical_kind)
    @skipIfEngineNotAvailable("tarski_grounder")
    def test_robot_loader(self):
        problem, plan = self.problems["robot_loader"]
        with Compiler(name="tarski_grounder") as grounder:
            ground_result = grounder.compile(problem, CompilationKind.GROUNDING)
            grounded_problem, rewrite_back_plan_function = (
                ground_result.problem,
                ground_result.map_back_action_instance,
            )
            for grounded_action in grounded_problem.actions:
                self.assertEqual(len(grounded_action.parameters), 0)
            with OneshotPlanner(problem_kind=grounded_problem.kind) as planner:
                self.assertNotEqual(planner, None)
                grounded_plan = planner.solve(grounded_problem).plan
                plan = grounded_plan.replace_action_instances(
                    rewrite_back_plan_function
                )
                for ai in plan.actions:
                    a = ai.action
                    self.assertEqual(a, problem.action(a.name))
                with PlanValidator(
                    problem_kind=problem.kind, plan_kind=plan.kind
                ) as pv:
                    self.assertTrue(pv.validate(problem, plan))

    @skipIfNoOneshotPlannerForProblemKind(full_classical_kind)
    @skipIfNoPlanValidatorForProblemKind(full_classical_kind)
    @skipIfEngineNotAvailable("tarski_grounder")
    def test_robot_locations_connected_without_battery(self):
        problem, plan = self.problems["robot_locations_connected_without_battery"]
        with Compiler(name="tarski_grounder") as grounder:
            ground_result = grounder.compile(problem, CompilationKind.GROUNDING)
            grounded_problem, rewrite_back_plan_function = (
                ground_result.problem,
                ground_result.map_back_action_instance,
            )
            for grounded_action in grounded_problem.actions:
                self.assertEqual(len(grounded_action.parameters), 0)
            with OneshotPlanner(problem_kind=grounded_problem.kind) as planner:
                self.assertNotEqual(planner, None)
                grounded_plan = planner.solve(grounded_problem).plan
                plan = grounded_plan.replace_action_instances(
                    rewrite_back_plan_function
                )
                for ai in plan.actions:
                    a = ai.action
                    self.assertEqual(a, problem.action(a.name))
                with PlanValidator(
                    problem_kind=problem.kind, plan_kind=plan.kind
                ) as pv:
                    self.assertTrue(pv.validate(problem, plan))

    @skipIfNoOneshotPlannerForProblemKind(basic_classical_kind.union(hierarchical_kind))
    @skipIfNoPlanValidatorForProblemKind(basic_classical_kind.union(hierarchical_kind))
    @skipIfEngineNotAvailable("tarski_grounder")
    def test_hierarchical_blocks_world(self):
        problem, plan = self.problems["hierarchical_blocks_world"]
        with Compiler(name="tarski_grounder") as grounder:
            self.assertTrue(grounder.is_compiler())
            self.assertTrue(grounder.supports(problem.kind))
            self.assertTrue(grounder.supports_compilation(CompilationKind.GROUNDING))
            ground_result = grounder.compile(problem, CompilationKind.GROUNDING)
            grounded_problem, rewrite_back_plan_function = (
                ground_result.problem,
                ground_result.map_back_action_instance,
            )
            for grounded_action in grounded_problem.actions:
                self.assertEqual(len(grounded_action.parameters), 0)
            with OneshotPlanner(problem_kind=grounded_problem.kind) as planner:
                self.assertNotEqual(planner, None)
                grounded_plan = planner.solve(grounded_problem).plan
                plan = grounded_plan.replace_action_instances(
                    rewrite_back_plan_function
                )
                for ai in plan.actions:
                    a = ai.action
                    self.assertEqual(a, problem.action(a.name))
                with PlanValidator(
                    problem_kind=problem.kind, plan_kind=plan.kind
                ) as pv:
                    self.assertTrue(pv.validate(problem, plan))

    @skipIfNoOneshotPlannerForProblemKind(basic_classical_kind)
    @skipIfEngineNotAvailable("tarski_grounder")
    def test_tarski_grounder_mockup_problem(self):
        problem = Problem("mockup")
        Location = UserType("Location")
        at = Fluent("at", BoolType(), position=Location)
        at_l1 = Fluent("at_l1")
        at_l2 = Fluent("at_l2")
        l1 = Object("l1", Location)
        l2 = Object("l2", Location)
        move_to = InstantaneousAction("move_to", l_to=Location)
        l_to = move_to.parameter("l_to")
        move_to.add_effect(at(l_to), True)
        move_to.add_effect(at_l1, False, Equals(l_to, l2))
        move_to_l2 = InstantaneousAction("move_to_l2")
        move_to_l2.add_effect(at_l2, True)
        problem.add_fluent(at, default_initial_value=False)
        problem.add_fluent(at_l1, default_initial_value=True)
        problem.add_fluent(at_l2, default_initial_value=False)
        problem.add_object(l1)
        problem.add_object(l2)
        problem.add_action(move_to)
        problem.add_action(move_to_l2)
        problem.add_goal(at(l1))
        problem.add_goal(at(l2))
        problem.add_goal(at_l2)

        with Compiler(name="tarski_grounder") as grounder:
            ground_result = grounder.compile(problem, CompilationKind.GROUNDING)
            grounded_problem, rewrite_back_plan_function = (
                ground_result.problem,
                ground_result.map_back_action_instance,
            )
            for grounded_action in grounded_problem.actions:
                self.assertEqual(len(grounded_action.parameters), 0)
            with OneshotPlanner(problem_kind=grounded_problem.kind) as planner:
                self.assertNotEqual(planner, None)
                grounded_plan = planner.solve(grounded_problem).plan
                plan = grounded_plan.replace_action_instances(
                    rewrite_back_plan_function
                )
                for ai in plan.actions:
                    a = ai.action
                    self.assertEqual(a, problem.action(a.name))
                with PlanValidator(
                    problem_kind=problem.kind, plan_kind=plan.kind
                ) as pv:
                    self.assertTrue(pv.validate(problem, plan))



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unified_planning
from unified_planning.shortcuts import *
from unified_planning.test import TestCase, main
from unified_planning.model.walkers import Simplifier, Substituter
from unified_planning.environment import get_env
from fractions import Fraction


class TestBoolOperators(TestCase):
    def setUp(self):
        TestCase.setUp(self)

    def test_and_constant(self):
        s = Simplifier(get_env())
        t = Bool(True)
        f = Bool(False)
        e1 = And(t, f)
        r1 = s.simplify(e1)
        self.assertEqual(r1, f)
        self.assertEqual(r1.constant_value(), False)
        e2 = And(t, e1)
        r2 = s.simplify(e2)
        self.assertEqual(r2, f)
        self.assertEqual(r2.constant_value(), False)
        e3 = And(t, t)
        r3 = s.simplify(e3)
        self.assertTrue(r3.is_constant())
        self.assertEqual(r3, t)

    def test_and_fluent(self):
        s = Simplifier(get_env())
        x = Fluent("x")
        y = Fluent("y")
        t = Bool(True)
        f = Bool(False)
        e1 = And(x, f)
        r1 = s.simplify(e1)
        self.assertEqual(r1, f)
        self.assertEqual(r1.constant_value(), False)
        e2 = And(x, e1)
        r2 = s.simplify(e2)
        self.assertEqual(r2, f)
        self.assertEqual(r2.constant_value(), False)
        e3 = And(x, t, t)
        e4 = And(e3, t)
        r4 = s.simplify(e4)
        self.assertEqual(FluentExp(x), r4)
        e5 = And(x, x)
        r5 = s.simplify(e5)
        self.assertEqual(r5, FluentExp(x))
        e6 = And(And(x, y), And(y, x))
        r6 = s.simplify(e6)
        self.assertEqual(r6, And(x, y))
        e7 = And(And(x, y), And(y, Not(x)))
        r7 = s.simplify(e7)
        self.assertEqual(r7, f)
        e8 = And(And(x, y), Not(x))
        r8 = s.simplify(e8)
        self.assertEqual(r8, f)

    def test_or_constant(self):
        s = Simplifier(get_env())
        t = Bool(True)
        f = Bool(False)
        e1 = Or(t, f)
        r1 = s.simplify(e1)
        self.assertEqual(r1, t)
        self.assertEqual(r1.constant_value(), True)
        e2 = Or(t, e1)
        r2 = s.simplify(e2)
        self.assertEqual(r2, t)
        self.assertEqual(r2.constant_value(), True)
        e3 = Or(f, f)
        r3 = s.simplify(e3)
        self.assertTrue(r3.is_constant())
        self.assertEqual(r3, f)

    def test_or_fluent(self):
        s = Simplifier(get_env())
        x = Fluent("x")
        y = Fluent("y")
        t = Bool(True)
        f = Bool(False)
        e1 = Or(x, t)
        r1 = s.simplify(e1)
        self.assertEqual(r1, t)
        self.assertEqual(r1.constant_value(), True)
        e2 = Or(x, e1)
        r2 = s.simplify(e2)
        self.assertEqual(r2, t)
        self.assertEqual(r2.constant_value(), True)
        e3 = Or(x, f, f)
        e4 = Or(e3, f)
        r4 = s.simplify(e4)
        self.assertEqual(FluentExp(x), r4)
        e5 = Or(x, x)
        r5 = s.simplify(e5)
        self.assertEqual(r5, FluentExp(x))
        e6 = Or(Or(x, y), Or(y, x))
        r6 = s.simplify(e6)
        self.assertEqual(r6, Or(x, y))
        e7 = Or(Or(x, y), Or(y, Not(x)))
        r7 = s.simplify(e7)
        self.assertEqual(r7, t)
        e8 = Or(Or(x, y), Not(x))
        r8 = s.simplify(e8)
        self.assertEqual(r8, t)

    def test_not(self):
        s = Simplifier(get_env())
        x = Fluent("x")
        t = Bool(True)
        f = Bool(False)
        e1 = Or(x, Not(t))
        r1 = s.simplify(e1)
        self.assertEqual(r1, FluentExp(x))
        e2 = Not(Or(x, e1))
        r2 = s.simplify(e2)
        self.assertEqual(r2, Not(x))
        e3 = Or(x, f, f)
        e4 = Not(Or(e3, f))
        r4 = s.simplify(Not(e4))
        self.assertEqual(FluentExp(x), r4)
        e5 = Not(Not(x))
        r5 = s.simplify(e5)
        self.assertEqual(r5, FluentExp(x))

    def test_iff(self):
        s = Simplifier(get_env())
        x = FluentExp(Fluent("x"))
        y = FluentExp(Fluent("y"))
        t = Bool(True)
        f = Bool(False)
        e1 = Iff(x, y)
        r1 = s.simplify(e1)
        self.assertEqual(r1, e1)
        e2 = Iff(e1, e1)
        r2 = s.simplify(e2)
        self.assertEqual(r2, t)
        e3 = Iff(And(t, y), Or(y, f))
        r3 = s.simplify(e3)
        self.assertEqual(r3, t)
        e4 = Iff(And(x, y), And(y, x))
        r4 = s.simplify(e4)
        self.assertEqual(r4, e4)
        e5 = Iff(And(x, y), And(x, y))
        r5 = s.simplify(e5)
        self.assertEqual(r5, t)
        e6 = Iff(And(t, t), f)
        r6 = s.simplify(e6)
        self.assertEqual(r6, f)
        e7 = Iff(And(x, t), f)
        r7 = s.simplify(e7)
        self.assertEqual(r7, Not(x))
        e8 = Iff(And(x, t), t)
        r8 = s.simplify(e8)
        self.assertEqual(r8, x)
        e9 = Iff(f, And(x, t))
        r9 = s.simplify(e9)
        self.assertEqual(r9, Not(x))
        e10 = Iff(t, And(x, t))
        r10 = s.simplify(e10)
        self.assertEqual(r10, x)

    def test_implies(self):
        s = Simplifier(get_env())
        x = FluentExp(Fluent("x"))
        y = FluentExp(Fluent("y"))
        t = Bool(True)
        f = Bool(False)
        e1 = Implies(x, y)
        r1 = s.simplify(e1)
        self.assertEqual(r1, e1)
        e2 = Implies(e1, e1)
        r2 = s.simplify(e2)
        self.assertEqual(r2, t)
        e3 = Implies(And(t, y), Or(y, f))
        r3 = s.simplify(e3)
        self.assertEqual(r3, t)
        e4 = Implies(And(x, y), And(y, x))
        r4 = s.simplify(e4)
        self.assertEqual(r4, e4)
        e5 = Implies(And(x, y), y)
        r5 = s.simplify(e5)
        self.assertEqual(r5, e5)
        e6 = Implies(And(x, y), t)
        r6 = s.simplify(e6)
        self.assertEqual(r6, t)
        e7 = Implies(And(x, f), Not(f))
        r7 = s.simplify(e7)
        self.assertEqual(r7, t)
        e8 = Implies(t, x)
        r8 = s.simplify(e8)
        self.assertEqual(r8, x)
        e9 = Implies(x, f)
        r9 = s.simplify(e9)
        self.assertEqual(r9, Not(x))

    def test_equals(self):
        s = Simplifier(get_env())
        x = FluentExp(Fluent("x", IntType()))
        y = FluentExp(Fluent("y", IntType()))
        t = Bool(True)
        f = Bool(False)
        i1 = Int(5)
        i2 = Int(6)
        i3 = Real(Fraction(6))
        e1 = Equals(x, y)
        r1 = s.simplify(e1)
        self.assertEqual(r1, e1)
        e2 = Equals(i1, i2)
        r2 = s.simplify(e2)
        self.assertEqual(r2, f)
        e3 = Equals(i2, i3)
        r3 = s.simplify(e3)
        self.assertEqual(r3, t)
        e4 = Equals(x, x)
        r4 = s.simplify(e4)
        self.assertEqual(r4, t)

    def test_le(self):
        s = Simplifier(get_env())
        x = FluentExp(Fluent("x", IntType()))
        y = FluentExp(Fluent("y", IntType()))
        t = Bool(True)
        f = Bool(False)
        i1 = Int(5)
        i2 = Int(6)
        i3 = Real(Fraction(6))
        e1 = LE(x, y)
        r1 = s.simplify(e1)
        self.assertEqual(r1, e1)
        e2 = LE(i1, i2)
        r2 = s.simplify(e2)
        self.assertEqual(r2, t)
        e3 = LE(i2, i3)
        r3 = s.simplify(e3)
        self.assertEqual(r3, t)
        e4 = LE(i2, i1)
        r4 = s.simplify(e4)
        self.assertEqual(r4, f)

    def test_lt(self):
        s = Simplifier(get_env())
        x = FluentExp(Fluent("x", IntType()))
        y = FluentExp(Fluent("y", IntType()))
        t = Bool(True)
        f = Bool(False)
        i1 = Int(5)
        i2 = Int(6)
        i3 = Real(Fraction(6))
        e1 = LT(x, y)
        r1 = s.simplify(e1)
        self.assertEqual(r1, e1)
        e2 = LT(i1, i2)
        r2 = s.simplify(e2)
        self.assertEqual(r2, t)
        e3 = LT(i2, i3)
        r3 = s.simplify(e3)
        self.assertEqual(r3, f)
        e4 = LT(i2, i1)
        r4 = s.simplify(e4)
        self.assertEqual(r4, f)

    def test_gt(self):
        s = Simplifier(get_env())
        x = FluentExp(Fluent("x", IntType()))
        y = FluentExp(Fluent("y", IntType()))
        t = Bool(True)
        f = Bool(False)
        i1 = Int(5)
        i2 = Int(6)
        i3 = Real(Fraction(6))
        e1 = GT(x, y)
        r1 = s.simplify(e1)
        self.assertEqual(r1, e1)
        e2 = GT(i1, i2)
        r2 = s.simplify(e2)
        self.assertEqual(r2, f)
        e3 = GT(i2, i3)
        r3 = s.simplify(e3)
        self.assertEqual(r3, f)
        e4 = GT(i2, i1)
        r4 = s.simplify(e4)
        self.assertEqual(r4, t)

    def test_ge(self):
        s = Simplifier(get_env())
        x = FluentExp(Fluent("x", IntType()))
        y = FluentExp(Fluent("y", IntType()))
        t = Bool(True)
        f = Bool(False)
        i1 = Int(5)
        i2 = Int(6)
        i3 = Real(Fraction(6))
        e1 = GE(x, y)
        r1 = s.simplify(e1)
        self.assertEqual(r1, e1)
        e2 = GE(i1, i2)
        r2 = s.simplify(e2)
        self.assertEqual(r2, f)
        e3 = GE(i2, i3)
        r3 = s.simplify(e3)
        self.assertEqual(r3, t)
        e4 = GE(i2, i1)
        r4 = s.simplify(e4)
        self.assertEqual(r4, t)


class TestArithmeticOperators(TestCase):
    def setUp(self):
        TestCase.setUp(self)

    def test_plus_constant(self):
        # simple plus
        s = Simplifier(get_env())
        data1 = Fraction(5.0)
        data2 = 3
        fnode1 = Real(data1)
        fnode2 = Int(data2)
        fnode1_2 = Plus(fnode1, fnode2)
        result1_2 = s.simplify(fnode1_2)
        self.assertTrue(result1_2.constant_value() == 8)
        data_list = [1, 5, 6, 2, 3, 4, -3, 5]
        fnode_of_data_list = Plus(data_list)
        fnode_simplified = s.simplify(fnode_of_data_list)
        self.assertTrue(fnode_simplified.is_int_constant())
        self.assertTrue(fnode_simplified.constant_value() == 23)
        data_list = [1, 5, 6, 2, -10, -5, 3, -2]
        fnode_of_data_list = Plus(data_list)
        fnode_simplified = s.simplify(fnode_of_data_list)
        self.assertTrue(fnode_simplified.is_int_constant())
        self.assertEqual(fnode_simplified.constant_value(), 0)

    def test_plus_fluent(self):
        # plus with fluent
        s = Simplifier(get_env())
        data2 = 3
        x = Fluent("x", IntType())
        y = Fluent("y", IntType())
        fnode2 = Int(data2)
        fnodex_2 = Plus(x, fnode2)
        data_list = [1, 5, 6, 2, 3, 4, -3, 5]

        fnode_of_data_list = Plus(data_list)
        fnode_of_data_list = Plus(fnode_of_data_list, fnodex_2)
        fnode_simplified = s.simplify(fnode_of_data_list)
        self.assertEqual(fnode_simplified, Plus(x, Int(26)))

        data_list = [1, 5, 6, 2, -10, -5, 3, -2, x, y]
        fnode_of_data_list = Plus(data_list)
        fnode_simplified = s.simplify(fnode_of_data_list)
        self.assertEqual(fnode_simplified, Plus(x, y))

    def test_minus_constant(self):
        # simple minus
        s = Simplifier(get_env())
        data1 = 5
        data2 = 3
        fnode1 = Int(data1)
        fnode2 = Int(data2)
        fnode1_2 = Minus(fnode1, fnode2)
        result1_2 = s.simplify(fnode1_2)
        self.assertEqual(result1_2.constant_value(), 2)
        data_list = [1, 5, 6, 2, 3, 4, -3, 5]
        fnode_of_data_list = Int(data_list.pop(0))
        for a in data_list:
            fnode_of_data_list = Minus(fnode_of_data_list, Int(a))
        fnode_simplified = s.simplify(fnode_of_data_list)
        self.assertEqual(fnode_simplified.constant_value(), -21)

    def test_minus_fluent(self):
        # minus with fluent
        s = Simplifier(get_env())
        data2 = 3
        x = Fluent("x", IntType())
        y = Fluent("y", IntType())
        fnode2 = Int(data2)
        x_2 = Minus(x, fnode2)
        data_list = [1, 5, 6, 2, 3, 4, -3, 5]
        fnode_of_data_list = x_2
        for a in data_list:
            fnode_of_data_list = Minus(fnode_of_data_list, Int(a))
        fnode_simplified = s.simplify(fnode_of_data_list)
        fnode_of_data_list_expected = x_2
        for a in data_list:
            if a > 0:
                fnode_of_data_list_expected = Minus(fnode_of_data_list_expected, Int(a))
            else:
                fnode_of_data_list_expected = Plus(fnode_of_data_list_expected, Int(-a))
        self.assertEqual(fnode_simplified, fnode_of_data_list_expected)

        fnode_of_data_list = Int(0)
        for a in data_list:
            fnode_of_data_list = Minus(fnode_of_data_list, Int(a))
        x_fnode_of_data_list = Minus(x, fnode_of_data_list)
        fnode_simplified = s.simplify(x_fnode_of_data_list)
        self.assertEqual(fnode_simplified, Plus(x, Int(23)))

        e1 = Minus(x, y)
        r1 = s.simplify(e1)
        self.assertEqual(r1, e1)

    def test_times_constant(self):
        # simple times
        s = Simplifier(get_env())
        data1 = 5
        data2 = 3
        fnode1 = Int(data1)
        fnode2 = Int(data2)
        fnode1_2 = Times(fnode1, fnode2)
        result1_2 = s.simplify(fnode1_2)

        self.assertTrue(result1_2.constant_value() == 15)
        data_list = [1, 5, 6, 2, 3, 4, -3, 5]
        fnode_of_data_list = Times(0, *data_list)
        fnode_simplified = s.simplify(fnode_of_data_list)
        self.assertTrue(fnode_simplified.is_int_constant())
        self.assertEqual(fnode_simplified.constant_value(), 0)
        data_list = [1, 5, 6, 2, 3, 4, -3, 5]
        fnode_of_data_list = Times(data_list)
        fnode_simplified = s.simplify(fnode_of_data_list)
        self.assertTrue(fnode_simplified.is_int_constant())
        self.assertEqual(fnode_simplified.constant_value(), -10800)

        e1 = Times(Int(5), Minus(-3, Plus(-5, 2)))
        r1 = s.simplify(e1)
        self.assertTrue(r1.is_constant())
        self.assertTrue(r1.constant_value() == 0)
        self.assertEqual(r1, Int(0))

    def test_times_fluent(self):
        # plus with fluent
        s = Simplifier(get_env())
        data2 = 3
        x = FluentExp(Fluent("x", IntType()))
        fnode2 = Int(data2)
        x_2 = Times(x, fnode2)
        data_list = [1, 5, 6, 2, 3, 4, -3, 5]

        fnode_of_data_list = Int(0)
        for a in data_list:
            fnode_of_data_list = Times(fnode_of_data_list, Int(a))
        fnode_of_data_list = Times(fnode_of_data_list, x_2)
        fnode_simplified = s.simplify(fnode_of_data_list)
        self.assertEqual(fnode_simplified, Int(0))

        fnode_of_data_list = Int(1)
        for a in data_list:
            fnode_of_data_list = Times(fnode_of_data_list, Int(a))
        fnode_of_data_list = Times(fnode_of_data_list, x_2)
        fnode_simplified = s.simplify(fnode_of_data_list)
        self.assertEqual(fnode_simplified, Times(x, Int(-32400)))

        e1 = Times(x, Div(1, 5), Int(5))
        r1 = s.simplify(e1)
        self.assertEqual(r1, x)

    def test_div_constant(self):
        # simple div
        s = Simplifier(get_env())
        data1 = 5
        data2 = 3
        fnode1 = Int(data1)
        fnode2 = Int(data2)
        fnode1_2 = Div(fnode1, fnode2)
        result1_2 = s.simplify(fnode1_2)
        self.assertTrue(result1_2.is_constant())
        self.assertTrue(result1_2.constant_value() == Fraction(5, 3))
        data1 = 6
        data2 = 3
        fnode1 = Int(data1)
        fnode2 = Int(data2)
        fnode1_2 = Div(fnode1, fnode2)
        result1_2 = s.simplify(fnode1_2)
        self.assertTrue(result1_2.is_constant())
        self.assertTrue(result1_2.is_int_constant())
        self.assertTrue(result1_2.constant_value() == Fraction(2))
        self.assertTrue(result1_2 == Int(2))
        data1 = 6
        data2 = 4
        fnode1 = Int(data1)
        fnode2 = Int(data2)
        fnode1_2 = Div(fnode1, fnode2)
        result1_2 = s.simplify(fnode1_2)
        self.assertTrue(result1_2.is_constant())
        self.assertTrue(result1_2.is_real_constant())
        self.assertTrue(result1_2.constant_value() == Fraction(6, 4))
        self.assertTrue(result1_2 == Real(Fraction(6, 4)))

        data_list = ["1.0", "0.5", "10", "0.25", "0.125"]
        fnode_of_data_list = Int(125)
        for a in data_list:
            fnode_of_data_list = Div(fnode_of_data_list, Real(Fraction(a)))
        fnode_simplified = s.simplify(fnode_of_data_list)
        self.assertTrue(fnode_simplified.is_real_constant())
        self.assertEqual(fnode_simplified.constant_value(), 800)

    def test_div_fluent(self):
        # div with fluent
        s = Simplifier(get_env())
        data2 = 3
        x = Fluent("x", IntType())
        fnode2 = Int(data2)
        x_2 = Div(x, fnode2)
        data_list = [Fraction(5), Fraction(1, 5)]

        fnode_of_data_list = Int(1)
        for a in data_list:
            fnode_of_data_list = Div(fnode_of_data_list, Real(Fraction(a)))
        fnode_of_data_list = Div(x_2, fnode_of_data_list)
        fnode_simplified = s.simplify(fnode_of_data_list)
        self.assertEqual(fnode_simplified, Div(Div(x, Int(3)), Real(Fraction(1))))

        data_list = ["1.0", "0.5", "10", "0.25", "0.125"]
        fnode_of_data_list = Int(1)
        for a in data_list:
            fnode_of_data_list = Div(fnode_of_data_list, Real(Fraction(a)))
        fnode_of_data_list = Div(fnode_of_data_list, x_2)
        fnode_simplified = s.simplify(fnode_of_data_list)
        self.assertEqual(fnode_simplified, Div(Fraction("6.4"), Div(x, Int(3))))

    def test_general(self):
        s = Simplifier(get_env())
        x = FluentExp(Fluent("x"))
        y = FluentExp(Fluent("y", IntType()))
        t = Bool(True)
        f = Bool(False)
        # ((25/5)*30*2*2) - (20*5) (500) == (25*4*10) / 2 (500)
        e1 = Equals(
            Minus(Times([Div(25, 5), 30, 2, 2]), Times(20, 5)), Div(Times(25, 4, 10), 2)
        )
        r1 = s.simplify(e1)
        self.assertEqual(r1, t)
        # T => !x
        e2 = Implies(e1, Not(x))
        r2 = s.simplify(e2)
        self.assertEqual(r2, Not(x))
        # !x || (T => x)
        e3 = Or(e2, Implies(e1, x))
        r3 = s.simplify(e3)
        self.assertEqual(r3, t)
        self.assertEqual(r3, e3.simplify())


class TestWithSubstituter(TestCase):
    def setUp(self):
        TestCase.setUp(self)

    def test_and_fluent(self):
        s = Simplifier(get_env())
        su = Substituter(get_env())
        x = Fluent("x")
        y = FluentExp(Fluent("y"))
        t = Bool(True)
        f = Bool(False)
        e1 = And(x, f)
        sub = {x: True}
        s1 = su.substitute(e1, sub)
        r1 = s.simplify(s1)
        self.assertEqual(r1, f)
        self.assertEqual(r1.constant_value(), False)
        e2 = And(x, e1)
        sub = {x: True}
        s2 = su.substitute(e2, sub)
        r2 = s.simplify(s2)
        self.assertEqual(r2, f)
        self.assertEqual(r2.constant_value(), False)
        e3 = And(x, t, t)
        e4 = And(e3, t)
        sub = {x: True}
        s4 = su.substitute(e4, sub)
        r4 = s.simplify(s4)
        self.assertEqual(t, r4)
        e5 = And(x, x)
        sub = {x: True}
        s5 = su.substitute(e5, sub)
        r5 = s.simplify(s5)
        self.assertEqual(r5, t)
        e6 = And(And(x, y), And(y, x))
        sub = {x: True}
        s6_1 = su.substitute(e6, sub)
        r6_1 = s.simplify(s6_1)
        self.assertEqual(r6_1, y)
        sub = {x: True, y: True}
        s6_2 = su.substitute(e6, sub)
        r6_2 = s.simplify(s6_2)
        self.assertEqual(r6_2, t)
        e7 = And(And(x, y), And(y, Not(x)))
        sub = {x: True, y: True}
        s7 = su.substitute(e7, sub)
        r7 = s.simplify(s7)
        self.assertEqual(r7, f)

    def test_or_fluent(self):
        s = Simplifier(get_env())
        su = Substituter(get_env())
        x = Fluent("x")
        y = FluentExp(Fluent("y"))
        t = Bool(True)
        f = Bool(False)
        e1 = Or(x, t)
        sub = {x: False}
        s1 = su.substitute(e1, sub)
        r1 = s.simplify(s1)
        self.assertEqual(r1, t)
        self.assertEqual(r1.constant_value(), True)
        e2 = Or(x, e1)
        sub = {x: False}
        s2 = su.substitute(e2, sub)
        r2 = s.simplify(s2)
        self.assertEqual(r2, t)
        self.assertEqual(r2.constant_value(), True)
        e3 = Or(x, f, f)
        e4 = Or(e3, f)
        sub = {x: False}
        s4_1 = su.substitute(e4, sub)
        r4_1 = s.simplify(s4_1)
        self.assertEqual(r4_1, f)
        sub = {x: True}
        s4_2 = su.substitute(e4, sub)
        r4_2 = s.simplify(s4_2)
        self.assertEqual(r4_2, t)
        e5 = Or(x, x)
        sub = {x: False}
        s5 = su.substitute(e5, sub)
        r5 = s.simplify(s5)
        self.assertEqual(r5, f)
        e6 = Or(Or(x, y), Or(y, x))
        sub = {x: False}
        s6 = su.substitute(e6, sub)
        r6 = s.simplify(s6)
        self.assertEqual(r6, y)
        e7 = Or(Or(x, y), Or(y, Not(x)))
        sub = {y: False}
        s7 = su.substitute(e7, sub)
        r7 = s.simplify(s7)
        self.assertEqual(r7, t)
        e8 = Or(Or(x, y), Not(x))
        sub = {y: False}
        s8 = su.substitute(e8, sub)
        r8 = s.simplify(s8)
        self.assertEqual(r8, t)
        sub = {x: False}
        s9 = su.substitute(e7, sub)
        r9 = s.simplify(s9)
        self.assertEqual(r9, t)

    def test_not(self):
        s = Simplifier(get_env())
        su = Substituter(get_env())
        x = Fluent("x")
        t = Bool(True)
        f = Bool(False)
        e1 = Or(x, Not(t))
        sub = {x: False}
        s1 = su.substitute(e1, sub)
        r1 = s.simplify(s1)
        self.assertEqual(r1, f)
        e2 = Not(Or(x, e1))
        sub = {x: False}
        s2 = su.substitute(e2, sub)
        r2 = s.simplify(s2)
        self.assertEqual(r2, t)

    def test_iff(self):
        s = Simplifier(get_env())
        su = Substituter(get_env())
        x = FluentExp(Fluent("x"))
        y = FluentExp(Fluent("y"))
        t = Bool(True)
        f = Bool(False)
        e1 = Iff(x, y)
        sub = {x: False}
        s1 = su.substitute(e1, sub)
        r1 = s.simplify(s1)
        self.assertEqual(r1, Not(y))
        e4 = Iff(And(x, y), And(y, x))
        sub = {x: False}
        s4 = su.substitute(e4, sub)
        r4 = s.simplify(s4)
        self.assertEqual(r4, t)
        e5 = Iff(And(x, t), f)
        sub = {x: False}
        s5 = su.substitute(e5, sub)
        r5 = s.simplify(s5)
        self.assertEqual(r5, t)

    def test_implies(self):
        s = Simplifier(get_env())
        su = Substituter(get_env())
        x = FluentExp(Fluent("x"))
        y = FluentExp(Fluent("y"))
        t = Bool(True)
        f = Bool(False)
        e1 = Implies(x, y)
        sub = {x: False}
        s1 = su.substitute(e1, sub)
        r1 = s.simplify(s1)
        self.assertEqual(r1, t)
        e3 = Implies(And(t, y), x)
        sub = {y: x, x: False}
        s3 = su.substitute(e3, sub)
        r3 = s.simplify(s3)
        self.assertEqual(r3, Not(x))

    def test_num_to_bools_operators(self):
        s = Simplifier(get_env())
        su = Substituter(get_env())
        a = FluentExp(Fluent("a", IntType()))
        b = FluentExp(Fluent("b", IntType()))
        c = FluentExp(Fluent("c", IntType()))
        d = FluentExp(Fluent("d", IntType()))
        x = FluentExp(Fluent("x"))
        y = FluentExp(Fluent("y"))
        t = Bool(True)
        f = Bool(False)
        # [(a+5 > b iff c != d) & x] iff (x xor y)
        e = Iff(
            And(x, Iff(GT(Plus(a, 5), b), Not(Equals(c, d)))),
            Or(And(x, Not(y)), And(Not(x), y)),
        )
        sub1 = {x: False, y: False, a: 0, b: 5, c: 0, d: 1}
        s1 = su.substitute(e, sub1)
        r1 = s.simplify(s1)
        self.assertEqual(r1, t)
        sub2 = {x: True, y: False, a: 0, b: 5, c: 0, d: 1}
        s2 = su.substitute(e, sub2)
        r2 = s.simplify(s2)
        self.assertEqual(r2, f)
        sub3 = {y: False, a: 0, b: 5, c: 0, d: 1}
        s3 = su.substitute(e, sub3)
        r3 = s.simplify(s3)
        self.assertEqual(r3, Not(x))
        sub4 = {y: False, a: 1, b: 5, c: 0, d: 1}
        s4 = su.substitute(e, sub4)
        r4 = s.simplify(s4)
        self.assertEqual(r4, t)
        sub5 = {Not(Equals(c, d)): GT(Plus(a, 5), b), y: False}
        s5 = su.substitute(e, sub5)
        r5 = s.simplify(s5)
        self.assertEqual(r5, t)


if __name__ == "__main__":
    main()



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from fractions import Fraction
from itertools import count
import re
import sys
from unicodedata import name
import unified_planning as up
import unified_planning.model.htn
import unified_planning.model.walkers as walkers
from unified_planning.model.types import _UserType, _IntType, _RealType
from typing import Dict, IO, Optional, cast
from io import StringIO


class ConverterToPythonString(walkers.DagWalker):
    """Expression converter to a Python string."""

    def __init__(
        self, env: "up.environment.Environment", names_mapping: Dict[str, str]
    ):
        walkers.DagWalker.__init__(self)
        self._names_mapping = names_mapping

    def convert(self, expression):
        """Converts the given expression to a python command."""
        if expression is None:
            return "None"
        return self.walk(expression)

    def walk_exists(self, expression, args):
        assert len(args) == 1
        vars_string_list = [
            f'up.model.Variable("{v.name}", {_print_python_type(v.type, self._names_mapping)})'
            for v in expression.variables()
        ]
        return f'emgr.Exists({args[0]}, {", ".join(vars_string_list)})'

    def walk_forall(self, expression, args):
        assert len(args) == 1
        vars_string_list = [
            f'up.model.Variable("{v.name}", {_print_python_type(v.type, self._names_mapping)})'
            for v in expression.variables()
        ]
        return f'emgr.Forall({args[0]}, {", ".join(vars_string_list)})'

    def walk_variable_exp(self, expression, args):
        assert len(args) == 0
        v = expression.variable()
        return f'emgr.VariableExp(up.model.Variable("{v.name}", {_print_python_type(v.type, self._names_mapping)}))'

    def walk_and(self, expression, args):
        assert len(args) > 1
        return f'emgr.And({", ".join(args)})'

    def walk_or(self, expression, args):
        assert len(args) > 1
        return f'emgr.Or({", ".join(args)})'

    def walk_not(self, expression, args):
        assert len(args) == 1
        return f"emgr.Not({args[0]})"

    def walk_implies(self, expression, args):
        assert len(args) == 2
        return f"emgr.Implies({args[0]}, {args[1]})"

    def walk_iff(self, expression, args):
        assert len(args) == 2
        return f"emgr.Iff({args[0]}, {args[1]})"

    def walk_fluent_exp(self, expression, args):
        fluent = expression.fluent()
        if args:
            return f'{_get_mangled_name(f"fluent_{fluent.name}", self._names_mapping)}({", ".join(args)})'
        return f'emgr.FluentExp({_get_mangled_name(f"fluent_{fluent.name}", self._names_mapping)})'

    def walk_param_exp(self, expression, args):
        assert len(args) == 0
        p = expression.parameter()
        return f'emgr.ParameterExp(up.model.Parameter("{p.name}", {_print_python_type(p.type, self._names_mapping)}, env))'

    def walk_object_exp(self, expression, args):
        assert len(args) == 0
        o = expression.object()
        return f'emgr.ObjectExp({_get_mangled_name(f"object_{o.name}", self._names_mapping)})'

    def walk_timing_exp(self, expression, args):
        assert len(args) == 0
        return _convert_timing(expression.timing())

    def walk_bool_constant(self, expression, args):
        assert len(args) == 0
        if expression.bool_constant_value():
            return "emgr.TRUE()"
        else:
            return "emgr.FALSE()"

    def walk_real_constant(self, expression, args):
        assert len(args) == 0
        return f"emgr.Real(Fraction({str(expression.real_constant_value().numerator)}, {str(expression.real_constant_value().denominator)}))"

    def walk_int_constant(self, expression, args):
        assert len(args) == 0
        return f"emgr.Int({str(expression.constant_value())})"

    def walk_plus(self, expression, args):
        assert len(args) > 1
        return f'emgr.Plus({", ".join(args)})'

    def walk_minus(self, expression, args):
        assert len(args) == 2
        return f"emgr.Minus({args[0]}, {args[1]})"

    def walk_times(self, expression, args):
        assert len(args) > 1
        return f'emgr.Times({", ".join(args)})'

    def walk_div(self, expression, args):
        assert len(args) == 2
        return f"emgr.Div({args[0]}, {args[1]})"

    def walk_le(self, expression, args):
        assert len(args) == 2
        return f"emgr.LE({args[0]}, {args[1]})"

    def walk_lt(self, expression, args):
        assert len(args) == 2
        return f"emgr.LT({args[0]}, {args[1]})"

    def walk_equals(self, expression, args):
        assert len(args) == 2
        return f"emgr.Equals({args[0]}, {args[1]})"


class PythonWriter:
    """This class can be used to write a Problem in PDDL."""

    def __init__(self, problem: "up.model.Problem"):
        self.problem = problem

    def _write_problem_code(self, out: IO[str]):

        names_mapping: Dict[str, str] = {}

        def params_as_dict(parameters) -> str:
            """Transforms a list of Parameter into an OrderedDict"""
            params = ", ".join(
                f'("{p.name}", {_print_python_type(p.type, names_mapping)})'
                for p in parameters
            )
            return f"OrderedDict([{params}])"

        # Initialize names_mapping with valid names
        for type in self.problem.user_types:  # define user_types
            utype = cast(_UserType, type)
            type_var_name = f"type_{utype.name}"
            if type_var_name.isidentifier():
                names_mapping[type_var_name] = type_var_name
        for fluent in self.problem.fluents:
            fluent_var_name = f"fluent_{fluent.name}"
            if fluent_var_name.isidentifier():
                names_mapping[fluent_var_name] = fluent_var_name
        for object in self.problem.all_objects:
            object_var_name = f"object_{object.name}"
            if object_var_name.isidentifier():
                names_mapping[object_var_name] = object_var_name
        for action in self.problem.actions:
            action_var_name = f"act_{action.name}"
            if action_var_name.isidentifier():
                names_mapping[action_var_name] = action_var_name

        converter = ConverterToPythonString(self.problem.env, names_mapping)
        out.write("from fractions import Fraction\n")
        out.write("from collections import OrderedDict\n")
        out.write("import unified_planning as up\n")
        out.write("env = up.environment.get_env()\n")
        out.write("emgr = env.expression_manager\n")
        out.write("tm = env.type_manager\n")

        for type in self.problem.user_types:  # define user_types
            utype = cast(_UserType, type)
            if utype.father is None:
                out.write(
                    f'{_get_mangled_name(f"type_{utype.name}", names_mapping)} = tm.UserType("{utype.name}")\n'
                )
            else:
                out.write(
                    f'{_get_mangled_name(f"type_{utype.name}", names_mapping)} = tm.UserType("{utype.name}", {_get_mangled_name(f"type_{cast(_UserType, utype.father).name}", names_mapping)})\n'
                )

        for f in self.problem.fluents:  # define fluents
            params = ", ".join(
                f'("{p.name}", {_print_python_type(p.type, names_mapping)})'
                for p in f.signature
            )
            params = f"OrderedDict([{params}])"
            out.write(
                f'{_get_mangled_name(f"fluent_{f.name}", names_mapping)} = up.model.Fluent("{f.name}", {_print_python_type(f.type, names_mapping)}, _signature={params})\n'
            )

        for o in self.problem.all_objects:  # define objects
            out.write(
                f'{_get_mangled_name(f"object_{o.name}", names_mapping)} = up.model.Object("{o.name}", {_get_mangled_name(f"type_{cast(_UserType, o.type).name}", names_mapping)})\n'
            )

        out.write("problem_initial_defaults = {}\n")  # define initial_defaults
        for type, exp in self.problem.initial_defaults.items():
            out.write(
                f"problem_initial_defaults[{_print_python_type(type, names_mapping)}] = {converter.convert(exp)}\n"
            )

        if self.problem.kind.has_hierarchical():
            problem_class = "htn.HierarchicalProblem"
        else:
            problem_class = "Problem"
        problem_name = (
            f'"{self.problem.name}"' if self.problem.name is not None else "None"
        )
        out.write(
            f"problem = up.model.{problem_class}({problem_name}, env, initial_defaults=problem_initial_defaults)\n"
        )

        for o in self.problem.all_objects:  # add objects to the problem
            out.write(
                f'problem.add_object({_get_mangled_name(f"object_{o.name}", names_mapping)})\n'
            )

        for (
            f
        ) in (
            self.problem.fluents
        ):  # add fluents to the problem, with their fluents_default, if they have one
            default = self.problem.fluents_defaults.get(f, None)
            out.write(
                f'problem.add_fluent({_get_mangled_name(f"fluent_{f.name}", names_mapping)}'
            )
            if default is not None:  # the fluent has a default value
                out.write(f", default_initial_value={converter.convert(default)}")
            out.write(")\n")

        for a in self.problem.actions:  # define actions and add them to the problem
            if isinstance(a, up.model.InstantaneousAction):
                params = params_as_dict(a.parameters)
                out.write(
                    f'{_get_mangled_name(f"act_{a.name}", names_mapping)} = up.model.InstantaneousAction("{a.name}", _parameters={params})\n'
                )
                for p in a.preconditions:
                    out.write(
                        f'{_get_mangled_name(f"act_{a.name}", names_mapping)}.add_precondition({converter.convert(p)})\n'
                    )
                for e in a.effects:
                    if e.is_increase():
                        out.write(
                            f'{_get_mangled_name(f"act_{a.name}", names_mapping)}.add_increase_effect(fluent={converter.convert(e.fluent)}, value={converter.convert(e.value)}, condition={converter.convert(e.condition)})\n'
                        )
                    elif e.is_decrease():
                        out.write(
                            f'{_get_mangled_name(f"act_{a.name}", names_mapping)}.add_decrease_effect(fluent={converter.convert(e.fluent)}, value={converter.convert(e.value)}, condition={converter.convert(e.condition)})\n'
                        )
                    else:
                        out.write(
                            f'{_get_mangled_name(f"act_{a.name}", names_mapping)}.add_effect(fluent={converter.convert(e.fluent)}, value={converter.convert(e.value)}, condition={converter.convert(e.condition)})\n'
                        )
            elif isinstance(a, up.model.DurativeAction):
                out.write(
                    f'{_get_mangled_name(f"act_{a.name}", names_mapping)} = up.model.DurativeAction("{a.name}"'
                )
                for ap in a.parameters:
                    out.write(
                        f", {ap.name}={_print_python_type(ap.type, names_mapping)}"
                    )
                out.write(")\n")
                for ap in a.parameters:
                    out.write(
                        f'parameter_{ap.name} = {_get_mangled_name(f"act_{a.name}", names_mapping)}.parameter("{ap.name}")\n'
                    )
                out.write(
                    f'{_get_mangled_name(f"act_{a.name}", names_mapping)}.set_duration_constraint({_convert_interval_duration(a.duration, converter)})\n'
                )
                for i, cl in a.conditions.items():
                    for c in cl:
                        out.write(
                            f'{_get_mangled_name(f"act_{a.name}", names_mapping)}.add_condition({_convert_interval(i)}, {converter.convert(c)})\n'
                        )
                for t, el in a.effects.items():
                    for e in el:
                        if e.is_increase():
                            out.write(
                                f'{_get_mangled_name(f"act_{a.name}", names_mapping)}.add_increase_effect(timing={_convert_timing(t)}, fluent={converter.convert(e.fluent)}, value={converter.convert(e.value)}, condition={converter.convert(e.condition)})\n'
                            )
                        elif e.is_decrease():
                            out.write(
                                f'{_get_mangled_name(f"act_{a.name}", names_mapping)}.add_decrease_effect(timing={_convert_timing(t)}, fluent={converter.convert(e.fluent)}, value={converter.convert(e.value)}, condition={converter.convert(e.condition)})\n'
                            )
                        else:
                            out.write(
                                f'{_get_mangled_name(f"act_{a.name}", names_mapping)}.add_effect(timing={_convert_timing(t)}, fluent={converter.convert(e.fluent)}, value={converter.convert(e.value)}, condition={converter.convert(e.condition)})\n'
                            )
            else:
                raise NotImplementedError
            out.write(
                f'problem.add_action({_get_mangled_name(f"act_{a.name}", names_mapping)})\n'
            )

        for (
            f_exp,
            v_exp,
        ) in (
            self.problem.explicit_initial_values.items()
        ):  # add only previously added initial values
            out.write(
                f"problem.set_initial_value({converter.convert(f_exp)}, {converter.convert(v_exp)})\n"
            )

        for t, el in self.problem.timed_effects.items():  # add timed effects
            for e in el:
                if e.is_increase():
                    out.write(
                        f"problem.add_increase_effect(timing={_convert_timing(t)}, fluent={converter.convert(e.fluent)}, value={converter.convert(e.value)}, condition={converter.convert(e.condition)})\n"
                    )
                elif e.is_decrease():
                    out.write(
                        f"problem.add_decrease_effect(timing={_convert_timing(t)}, fluent={converter.convert(e.fluent)}, value={converter.convert(e.value)}, condition={converter.convert(e.condition)})\n"
                    )
                else:
                    out.write(
                        f"problem.add_timed_effect(timing={_convert_timing(t)}, fluent={converter.convert(e.fluent)}, value={converter.convert(e.value)}, condition={converter.convert(e.condition)})\n"
                    )

        for i, gl in self.problem.timed_goals.items():  # add timed goals
            for g in gl:
                out.write(
                    f"problem.add_timed_goal(interval={_convert_interval(i)}, goal={converter.convert(g)})\n"
                )

        for g in self.problem.goals:  # add goals
            out.write(f"problem.add_goal(goal={converter.convert(g)})\n")

        for qm in self.problem.quality_metrics:  # adding metrics
            if isinstance(qm, up.model.metrics.MinimizeActionCosts):
                out.write("costs = {}\n")
                for a, c in qm.costs.items():
                    out.write(f"costs[act_{a.name}] = {converter.convert(c)}\n")
            elif isinstance(qm, up.model.metrics.Oversubscription):
                out.write("goals = {}\n")
                for goal, cost in qm.goals.items():
                    out.write(f"goals[{converter.convert(goal)}] = {cost}\n")
            out.write("problem.add_quality_metric(")
            if isinstance(qm, up.model.metrics.MinimizeActionCosts):
                out.write(f"up.model.metrics.MinimizeActionCosts(costs, {qm.default})")
            elif isinstance(qm, up.model.metrics.MinimizeSequentialPlanLength):
                out.write("up.model.metrics.MinimizeSequentialPlanLength()")
            elif isinstance(qm, up.model.metrics.MinimizeMakespan):
                out.write("up.model.metrics.MinimizeMakespan()")
            elif isinstance(qm, up.model.metrics.MinimizeExpressionOnFinalState):
                out.write(
                    f"up.model.metrics.MinimizeExpressionOnFinalState({converter.convert(qm.expression)})"
                )
            elif isinstance(qm, up.model.metrics.MaximizeExpressionOnFinalState):
                out.write(
                    f"up.model.metrics.MaximizeExpressionOnFinalState({converter.convert(qm.expression)})"
                )
            elif isinstance(qm, up.model.metrics.Oversubscription):
                out.write("up.model.metrics.Oversubscription(goals)")
            else:
                raise NotImplementedError
            out.write(")\n")

        if self.problem.kind.has_hierarchical():
            assert isinstance(
                self.problem, up.model.htn.hierarchical_problem.HierarchicalProblem
            )

            for task in self.problem.tasks:
                params = params_as_dict(task.parameters)
                out.write(
                    f'{_get_mangled_name(f"task_{task.name}", names_mapping)} = up.model.htn.Task("{task.name}", _parameters={params})\n'
                )
                out.write(
                    f'problem.add_task({_get_mangled_name(f"task_{task.name}", names_mapping)})\n'
                )

            for m in self.problem.methods:
                params = params_as_dict(m.parameters)
                out.write(
                    f'{_get_mangled_name(f"method_{m.name}", names_mapping)} = up.model.htn.Method("{m.name}", _parameters={params})\n'
                )
                for mp in m.parameters:
                    out.write(
                        f'parameter_{mp.name} = {_get_mangled_name(f"method_{m.name}", names_mapping)}.parameter("{mp.name}")\n'
                    )
                achieved_task_name = _get_mangled_name(
                    f"task_{m.achieved_task.task.name}", names_mapping
                )
                achieved_task_params = ", ".join(
                    [
                        _get_mangled_name(f"parameter_{p.name}", names_mapping)
                        for p in m.achieved_task.parameters
                    ]
                )

                out.write(
                    f'{_get_mangled_name(f"method_{m.name}", names_mapping)}.set_task({achieved_task_name}, {achieved_task_params})\n'
                )
                for p in m.preconditions:
                    out.write(
                        f'{_get_mangled_name(f"method_{m.name}", names_mapping)}.add_precondition({converter.convert(p)})\n'
                    )
                for c in m.constraints:
                    out.write(
                        f'{_get_mangled_name(f"method_{m.name}", names_mapping)}.add_constraint({converter.convert(c)})\n'
                    )
                for st in m.subtasks:
                    if isinstance(st.task, up.model.htn.task.Task):
                        head = _get_mangled_name(f"task_{st.task.name}", names_mapping)
                    else:
                        head = _get_mangled_name(f"act_{st.task.name}", names_mapping)
                    params = ", ".join([converter.convert(p) for p in st.parameters])
                    out.write(
                        f'{_get_mangled_name(f"method_{m.name}", names_mapping)}.add_subtask({head}, {params}, ident="{st.identifier}")\n'
                    )
                out.write(
                    f'problem.add_method({_get_mangled_name(f"method_{m.name}", names_mapping)})\n'
                )

            for v in self.problem.task_network.variables:
                out.write(
                    f'problem.task_network.add_variable("{v.name}", {_print_python_type(v.type, names_mapping)})\n'
                )
            for st in self.problem.task_network.subtasks:
                if isinstance(st.task, up.model.htn.task.Task):
                    head = _get_mangled_name(f"task_{st.task.name}", names_mapping)
                else:
                    head = _get_mangled_name(f"act_{st.task.name}", names_mapping)
                params = ", ".join([converter.convert(p) for p in st.parameters])
                out.write(
                    f'problem.task_network.add_subtask({head}, {params}, ident="{st.identifier}")\n'
                )
            for c in self.problem.task_network.constraints:
                out.write(
                    f"problem.task_network.add_constraint({converter.convert(c)})\n"
                )

    def print_problem_python_commands(self):
        """Prints the string representing all the necessary commands to recreate the problem."""
        self._write_problem_code(sys.stdout)

    def write_problem_code(self) -> str:
        """Returns the string representing all the necessary commands to recreate the problem."""
        out = StringIO()
        self._write_problem_code(out)
        return out.getvalue()

    def write_problem_code_to_file(self, filename: str):
        """Dumps to file the PDDL domain."""
        with open(filename, "w") as f:
            self._write_problem_code(f)


def _get_mangled_name(original_name: str, names_mapping: Dict[str, str]) -> str:
    """Important note: This method updates the names_mapping"""
    regex = re.compile("^[a-zA-Z_]+.*")
    assert re.match(regex, original_name) is not None
    new_name: Optional[str] = names_mapping.get(original_name, None)
    if new_name is None:  # The name is not in the map, so it must be added
        if not original_name.isidentifier():  # Make the name a python identifier
            new_name = re.sub("[^0-9a-zA-Z_]", "_", f"{original_name}")
        else:
            new_name = original_name
        test_name = new_name  # Init values
        count = 0
        while test_name in names_mapping.values():  # Loop until we find a fresh name
            test_name = f"{new_name}_{str(count)}"
            count += 1
        new_name = test_name
        names_mapping[
            original_name
        ] = new_name  # Once a fresh valid name is found, update the map.
    assert new_name is not None
    assert new_name.isidentifier()
    return new_name


def _print_python_type(
    type: "up.model.types.Type", names_mapping: Dict[str, str]
) -> str:
    """This method takes a type and returns how to use it from command line.
    For the user_types it assumes they have already been created and just refers to them."""
    if type.is_user_type():
        return (
            f'{_get_mangled_name(f"type_{cast(_UserType, type).name}", names_mapping)}'
        )
    elif type.is_bool_type():
        return "tm.BoolType()"
    elif type.is_int_type():
        itype = cast(_IntType, type)
        return f"tm.IntType({str(itype.lower_bound)}, {str(itype.upper_bound)})"
    elif type.is_real_type():
        rtype = cast(_RealType, type)
        return f"tm.RealType({str(rtype.lower_bound)}, {str(rtype.upper_bound)})"
    else:
        raise NotImplementedError


def _convert_timing(timing: up.model.Timing) -> str:
    delay: str = f"{str(timing.delay)}"
    if isinstance(timing.delay, Fraction):
        delay = (
            f"Fraction({str(timing.delay.numerator)}, {str(timing.delay.denominator)})"
        )

    tp = timing.timepoint
    if timing.is_global():
        assert tp.container is None
        if timing.is_from_start():
            return f"up.model.GlobalStartTiming({delay})"
        else:
            return f"up.model.GlobalEndTiming({delay})"
    else:
        container = f'"{tp.container}"' if tp.container is not None else "None"
        if timing.is_from_start():
            return f"up.model.StartTiming({delay}, container={container})"
        else:
            return f"up.model.EndTiming({delay}, container={container})"


def _convert_interval(interval: up.model.TimeInterval) -> str:
    interval_feature: str = "up.model.ClosedTimeInterval"
    if interval.is_left_open() and interval.is_right_open():
        interval_feature = "up.model.OpenTimeInterval"
    elif interval.is_left_open() and not interval.is_right_open():
        interval_feature = "up.model.LeftOpenTimeInterval"
    elif not interval.is_left_open() and interval.is_right_open():
        interval_feature = "up.model.RightOpenTimeInterval"
    return f"{interval_feature}({_convert_timing(interval.lower)}, {_convert_timing(interval.upper)})"


def _convert_interval_duration(
    interval: up.model.DurationInterval, converter: ConverterToPythonString
) -> str:
    interval_feature: str = "up.model.ClosedDurationInterval"
    if interval.is_left_open() and interval.is_right_open():
        interval_feature = "up.model.OpenDurationInterval"
    elif interval.is_left_open() and not interval.is_right_open():
        interval_feature = "up.model.LeftOpenDurationInterval"
    elif not interval.is_left_open() and interval.is_right_open():
        interval_feature = "up.model.RightOpenDurationInterval"
    return f"{interval_feature}({converter.convert(interval.lower)}, {converter.convert(interval.upper)})"



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from unified_planning.model.action import Action, InstantaneousAction, DurativeAction
from unified_planning.model.effect import Effect, SimulatedEffect, EffectKind
from unified_planning.model.expression import (
    BoolExpression,
    Expression,
    ExpressionManager,
)
from unified_planning.model.fnode import FNode
from unified_planning.model.fluent import Fluent
from unified_planning.model.object import Object
from unified_planning.model.operators import OperatorKind
from unified_planning.model.parameter import Parameter
from unified_planning.model.abstract_problem import AbstractProblem
from unified_planning.model.problem import Problem
from unified_planning.model.problem_kind import ProblemKind
from unified_planning.model.state import ROState, COWState, UPCOWState
from unified_planning.model.timing import (
    TimepointKind,
    Timing,
    StartTiming,
    EndTiming,
    GlobalStartTiming,
    GlobalEndTiming,
    DurationInterval,
    ClosedDurationInterval,
)
from unified_planning.model.timing import (
    FixedDuration,
    OpenDurationInterval,
    LeftOpenDurationInterval,
    RightOpenDurationInterval,
)
from unified_planning.model.timing import (
    TimeInterval,
    TimePointInterval,
    ClosedTimeInterval,
    OpenTimeInterval,
    LeftOpenTimeInterval,
    RightOpenTimeInterval,
)
from unified_planning.model.types import Type, TypeManager
from unified_planning.model.variable import Variable, FreeVarsOracle
import unified_planning.model.metrics

__all__ = [
    "Action",
    "InstantaneousAction",
    "DurativeAction",
    "Effect",
    "SimulatedEffect",
    "EffectKind",
    "BoolExpression",
    "Expression",
    "ExpressionManager",
    "FNode",
    "Fluent",
    "Object",
    "OperatorKind",
    "Parameter",
    "AbstractProblem",
    "Problem",
    "ProblemKind",
    "ROState",
    "COWState",
    "UPCOWState",
    "TimepointKind",
    "Timing",
    "StartTiming",
    "EndTiming",
    "GlobalStartTiming",
    "GlobalEndTiming",
    "DurationInterval",
    "ClosedDurationInterval",
    "FixedDuration",
    "OpenDurationInterval",
    "LeftOpenDurationInterval",
    "RightOpenDurationInterval",
    "TimeInterval",
    "TimePointInterval",
    "ClosedTimeInterval",
    "OpenTimeInterval",
    "LeftOpenTimeInterval",
    "RightOpenTimeInterval",
    "Type",
    "TypeManager",
    "Variable",
    "FreeVarsOracle",
]
